import string

vietnamese_re = r'([^0-9a-zA-Z_ỹỷỵỳựữửừứủụợỡởờớộỗổồốỏọịỉệễểềếẽẻẹặẵẳằắậẫẩầấảạươũĩđăýúùõôóòíìêéèãâàáÀÁÂÃÈÉÊÌÍÒÓÔÕÙÚÝĂĐĨŨƠƯẠẢẤẦẨẪẬẮẰẴẶẸẺẼẾỀỂỄỆỈỊỌỎỐỒỔỖỘỚỜỞỠỢỤỦỨỪỬỮỰỲỴỶỸ])'
vietnamese_for_date_re = r'([^0-9a-zA-Z_ỹỷỵỳựữửừứủụợỡởờớộỗổồốỏọịỉệễểềếẽẻẹặẵẳằắậẫẩầấảạươũĩđăýúùõôóòíìêéèãâàáÀÁÂÃÈÉÊÌÍÒÓÔÕÙÚÝĂĐĨŨƠƯẠẢẤẦẨẪẬẮẰẴẶẸẺẼẾỀỂỄỆỈỊỌỎỐỒỔỖỘỚỜỞỠỢỤỦỨỪỬỮỰỲỴỶỸ%\$])'
vietnamese_without_num_re = r'([^a-zA-Z_ỹỷỵỳựữửừứủụợỡởờớộỗổồốỏọịỉệễểềếẽẻẹặẵẳằắậẫẩầấảạươũĩđăýúùõôóòíìêéèãâàáÀÁÂÃÈÉÊÌÍÒÓÔÕÙÚÝĂĐĨŨƠƯẠẢẤẦẨẪẬẮẰẴẶẸẺẼẾỀỂỄỆỈỊỌỎỐỒỔỖỘỚỜỞỠỢỤỦỨỪỬỮỰỲỴỶỸ])'
vietnamese_set = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYỹỷỵỳựữửừứủụợỡởờớộỗổồốỏọịỉệễểềếẽẻẹặẵẳằắậẫẩầấảạươũĩđăýúùõôóòíìêéèãâàáÀÁÂÃÈÉÊÌÍÒÓÔÕÙÚÝĂĐĨŨƠƯẠẢẤẦẨẪẬẮẰẴẶẸẺẼẾỀỂỄỆỈỊỌỎỐỒỔỖỘỚỜỞỠỢỤỦỨỪỬỮỰỲỴỶỸ'

opening_brackets_and_punctutations_re = r'([\(\[\{])(\s)'
punctutations_re = r'(\s)([\.,\?\!\:\;\)\]\}])'
punctuations = set([i for i in string.punctuation] + ['“‘”’'])



import pytest
from vieneu_utils.phonemize_text import phonemize_with_dict, phonemize_batch

def test_phonemize_vietnamese():
    text = "Xin chào Việt Nam"
    # We don't check exact phonemes because they depend on espeak version,
    # but we check if it returns a non-empty string and doesn't crash.
    phonemes = phonemize_with_dict(text)
    assert isinstance(phonemes, str)
    assert len(phonemes) > 0

def test_phonemize_english_tag():
    text = "Học <en>machine learning</en> rất hay"
    phonemes = phonemize_with_dict(text)
    assert isinstance(phonemes, str)
    assert len(phonemes) > 0
    # Should contain phonemes for 'machine learning'
    # usually espeak uses symbols like məˈʃiːn
    assert any(c in phonemes for c in "əˈʃ")

def test_phonemize_with_custom_dict():
    custom_dict = {"robot": "ro-bot-phi-diệu"}
    text = "Tôi là robot"
    phonemes = phonemize_with_dict(text, phoneme_dict=custom_dict)
    assert "ro-bot-phi-diệu" in phonemes

def test_phonemize_batch_consistency():
    texts = ["Xin chào", "Việt Nam", "Chào <en>world</en>"]
    results = phonemize_batch(texts)

    assert len(results) == 3
    for res in results:
        assert isinstance(res, str)
        assert len(res) > 0

    # Check consistency with single call
    assert results[0] == phonemize_with_dict(texts[0])
    assert results[1] == phonemize_with_dict(texts[1])
    assert results[2] == phonemize_with_dict(texts[2])



import pytest
import os
import sys
from unittest.mock import patch, MagicMock

# Thêm src vào path
sys.path.insert(0, os.path.abspath(os.path.join(os.getcwd(), 'src')))

from vieneu_utils.phonemize_text import phonemize_batch, phonemize_with_dict

def test_phonemize_batch_deduplication():
    # Use 3 texts with overlapping words and English segments
    texts = [
        "Cái Bàn <en>world</en>",
        "Cái Bàn <en>world</en>",
        "Cái Ghế <en>world</en>"
    ]

    # Patch the actual espeak phonemize call
    with patch("vieneu_utils.phonemize_text.phonemize") as mock_phonemize:
        # Call 1: force_espeak (EN từ <en> tag) → ['world']
        # Call 2: global_unknown VI accented  → ['bàn', 'cái', 'ghế']
        mock_phonemize.side_effect = [
            ["w-o-r-l-d"],             # Result for force_espeak EN words
            ["ban", "kai", "ge"],      # Result for VI unknown words
        ]

        results = phonemize_batch(texts, phoneme_dict={})

        assert mock_phonemize.call_count == 2

        all_calls = mock_phonemize.call_args_list

        # Call 1 = force_espeak (en-us), Call 2 = vi unknown
        en_call_words = all_calls[0][0][0]
        vi_call_words = all_calls[1][0][0]

        assert len(en_call_words) == 1   # chỉ 'world'
        assert len(vi_call_words) == 3   # 'bàn', 'cái', 'ghế'

        assert "world" in [w.lower() for w in en_call_words]
        assert "cái" in [w.lower() for w in vi_call_words]
        
def test_phonemize_with_dict_caching():
    from vieneu_utils.phonemize_text import _phonemize_with_dict_cached
    text = "Câu này sẽ được cache"

    # Clear cache before test
    _phonemize_with_dict_cached.cache_clear()

    with patch("vieneu_utils.phonemize_text.phonemize_batch") as mock_batch:
        mock_batch.return_value = ["p-h-o-n-e-m-e-s"]

        # First call
        res1 = phonemize_with_dict(text)
        # Second call
        res2 = phonemize_with_dict(text)

        assert res1 == res2
        # Should only be called once due to LRU cache
        assert mock_batch.call_count == 1

def test_base_ref_phoneme_cache():
    # Only import if base exists, otherwise skip or mock
    try:
        from vieneu.base import BaseVieneuTTS
    except ImportError:
        pytest.skip("BaseVieneuTTS not found")

    class MockTTS(BaseVieneuTTS):
        def infer(self, text, **kwargs):
            return None
        def infer_batch(self, texts, **kwargs):
            return [self.infer(t, **kwargs) for t in texts]

    tts = MockTTS()
    ref_text = "Giọng đọc mẫu số 1"

    with patch("vieneu_utils.phonemize_text.phonemize_with_dict") as mock_phonemize:
        mock_phonemize.return_value = "cached-phonemes"

        p1 = tts.get_ref_phonemes(ref_text)
        p2 = tts.get_ref_phonemes(ref_text)

        assert p1 == p2
        assert mock_phonemize.call_count == 1
        assert ref_text in tts._ref_phoneme_cache



import unittest
import os
import sys

# Thêm src vào path để import module
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'src')))

from vieneu_utils.phonemize_text import phonemize_text

class TestBilingualPhonemize(unittest.TestCase):
    """
    Test suite kiểm tra logic phonemize song ngữ và cơ chế lan truyền (propagation).
    """

    def assert_propagation(self, natural_text, tagged_text):
        """Kiểm tra câu tự nhiên phải cho ra phoneme giống hệt câu có tag."""
        p_nat = phonemize_text(natural_text)
        p_tag = phonemize_text(tagged_text)
        self.assertEqual(p_nat, p_tag, f"\nFAIL:\n  Nat: {p_nat}\n  Tag: {p_tag}")

    def test_bridge_propagation(self):
        """Kiểm tra từ 'common' đi theo mỏ neo English."""
        # 'go to the' được kéo sang EN bởi 'market'
        self.assert_propagation(
            "Tôi muốn go to the market", 
            "Tôi muốn <en>go to the market</en>"
        )

    def test_punctuation_boundary(self):
        """Dấu câu (dấu chấm) phải ngăn cách ngữ cảnh."""
        # Chữ 'to' ở túi to (VI) không được bị 'market' (EN) ở câu trước kéo đi
        self.assert_propagation(
            "go to the market. Mua một cái túi to.", 
            "<en>go to the market</en>. Mua một cái túi to."
        )

    def test_closest_neighbor_priority(self):
        """Từ common đi theo mỏ neo gần nhất."""
        # 'to' kẹp giữa 'túi' (VI) và 'market' (EN), nhưng sát 'túi' hơn
        self.assert_propagation("cái túi to market", "cái túi to <en>market</en>")
        
        # 'go to' kẹp giữa 'muốn' (VI) và 'market' (EN), khoảng cách bằng nhau -> ưu tiên bên phải (EN)
        self.assert_propagation("muốn go to market", "muốn <en>go to market</en>")

    def test_common_word_disambiguation(self):
        """Kiểm tra các cặp từ đa nghĩa phổ biến."""
        # me (EN) vs me (VI)
        # Note: In natural text without EN anchors, common words might default to VI.
        # So we test with an explicit EN anchor to verify propagation.
        self.assert_propagation("hello give it to me", "hello <en>give it to me</en>")
        self.assert_propagation("ăn quả me", "ăn quả me")
        
        # no (EN) vs no (VI)
        self.assert_propagation("hello I say no", "hello <en>I say no</en>")
        self.assert_propagation("ăn cho no", "ăn cho no")
        
        # can (EN) vs can (VI)
        self.assert_propagation("I can do it", "<en>I can do it</en>")
        self.assert_propagation("can ngăn", "can ngăn")

if __name__ == '__main__':
    unittest.main()



import numpy as np
import pytest
from vieneu_utils.core_utils import split_text_into_chunks, join_audio_chunks

def test_split_text_into_chunks():
    text = "Đây là một câu ngắn. Đây là một câu dài hơn một chút để kiểm tra xem nó có bị chia ra không nếu chúng ta đặt giới hạn ký tự thấp."
    chunks = split_text_into_chunks(text, max_chars=50)
    assert len(chunks) > 1
    for chunk in chunks:
        assert len(chunk) <= 50

def test_split_text_paragraphs():
    text = "Đoạn 1.\n\nĐoạn 2."
    chunks = split_text_into_chunks(text, max_chars=100)
    assert len(chunks) == 2
    assert "Đoạn 1" in chunks[0]
    assert "Đoạn 2" in chunks[1]

def test_join_audio_chunks():
    sr = 16000
    chunk1 = np.ones(1600, dtype=np.float32) # 0.1s
    chunk2 = np.ones(1600, dtype=np.float32) * 0.5

    # Simple join
    joined = join_audio_chunks([chunk1, chunk2], sr)
    assert len(joined) == 3200
    assert np.all(joined[:1600] == 1.0)
    assert np.all(joined[1600:] == 0.5)

def test_join_audio_chunks_with_silence():
    sr = 16000
    chunk1 = np.ones(1600, dtype=np.float32)
    chunk2 = np.ones(1600, dtype=np.float32)

    # Join with 0.1s silence (1600 samples)
    joined = join_audio_chunks([chunk1, chunk2], sr, silence_p=0.1)
    assert len(joined) == 1600 + 1600 + 1600
    assert np.all(joined[1600:3200] == 0.0)



import re
from .num2vi import n2w
from .symbols import vietnamese_re, vietnamese_for_date_re

day_in_month = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
_date_seperator = r"(\/|-|\.)"
_short_date_seperator = r"(\/|-)"

# Compiled Regular Expressions
RE_FULL_DATE = re.compile(r"\b(\d{1,2})" + _date_seperator + r"(\d{1,2})" + _date_seperator + r"(\d{4})\b", re.IGNORECASE)
RE_DAY_MONTH = re.compile(r"\b(\d{1,2})" + _short_date_seperator + r"(\d{1,2})\b", re.IGNORECASE)
RE_MONTH_YEAR = re.compile(r"\b(\d{1,2})" + _date_seperator + r"(\d{4})\b", re.IGNORECASE)
RE_FULL_TIME = re.compile(r"\b(\d{1,2})(g|:|h)(\d{1,2})(p|:|m)(\d{1,2})(?:\s*(giây|s|g))?\b", re.IGNORECASE)
RE_TIME = re.compile(r"\b(\d{1,2})(g|:|h)(\d{1,2})(?:\s*(phút|p|m))?\b", re.IGNORECASE)
RE_REDUNDANT_NGAY = re.compile(r'\bngày\s+ngày\b', re.IGNORECASE)

def _is_valid_date(day, month):
    try:
        day, month = int(day), int(month)
        return 1 <= month <= 12 and 1 <= day <= day_in_month[month - 1]
    except (ValueError, IndexError): return False

def _expand_full_date(match):
    day, sep1, month, sep2, year = match.groups()
    if _is_valid_date(day, month):
        day = str(int(day))
        month = str(int(month))
        return f"ngày {n2w(day)} tháng {n2w(month)} năm {n2w(year)}"
    return match.group(0)

def _expand_day_month(match):
    day, sep, month = match.groups()
    if _is_valid_date(day, month):
        day = str(int(day))
        month = str(int(month))
        return f"ngày {n2w(day)} tháng {n2w(month)}"
    return match.group(0)

def _norm_time_part(s):
    return '0' if s == '00' else s

def _expand_time(match):
    h, sep, m, suffix = match.groups()
    if 0 <= int(h) < 24 and 0 <= int(m) < 60:
        return f"{n2w(_norm_time_part(h))} giờ {n2w(_norm_time_part(m))} phút"
    return match.group(0)

def normalize_date(text):
    text = RE_FULL_DATE.sub(_expand_full_date, text)
    text = RE_MONTH_YEAR.sub(
        lambda m: f"tháng {n2w(str(int(m.group(1))))} năm {n2w(m.group(3))}",
        text
    )
    text = RE_DAY_MONTH.sub(_expand_day_month, text)
    text = RE_REDUNDANT_NGAY.sub('ngày', text)
    text = re.sub(r'\btháng\s+tháng\b', 'tháng', text, flags=re.IGNORECASE)
    text = re.sub(r'\bnăm\s+năm\b', 'năm', text, flags=re.IGNORECASE)
    return text

def normalize_time(text):
    text = RE_FULL_TIME.sub(
        lambda m: f"{n2w(str(int(m.group(1))))} giờ {n2w(str(int(m.group(3))))} phút {n2w(str(int(m.group(5))))} giây",
        text
    )
    text = RE_TIME.sub(_expand_time, text)
    return text



from typing import List, Optional

units = {
    '0': 'không',
    '1': 'một',
    '2': 'hai',
    '3': 'ba',
    '4': 'bốn',
    '5': 'năm',
    '6': 'sáu',
    '7': 'bảy',
    '8': 'tám',
    '9': 'chín',
}

def chunks(lst: List, n: int) -> List[List]:
    """Split list into chunks of size n."""
    return [lst[i : i + n] for i in range(0, len(lst), n)]

def n2w_units(numbers: str) -> str:
    """Convert single digit to word."""
    if not numbers:
        raise ValueError('Số rỗng!!')
    if len(numbers) > 1:
        raise ValueError('Số vượt quá giá trị của hàng đơn vị!')
    return units[numbers]

def pre_process_n2w(number: str) -> Optional[str]:
    """Clean input number string."""
    clean = str(number).translate(str.maketrans('', '', ' -.,'))
    return clean if clean.isdigit() else None

def process_n2w_single(numbers: str) -> str:
    """Convert each digit independently to word."""
    return ' '.join(units[d] for d in numbers if d in units)

def n2w_hundreds(numbers: str) -> str:
    """Convert 1-3 digit number to Vietnamese words."""
    if not numbers or numbers == '000':
        return ""
    
    # Pad to 3 digits
    n = numbers.zfill(3)
    h_digit, t_digit, u_digit = n[0], n[1], n[2]

    res = []

    # Hundreds
    if h_digit != '0':
        res.append(units[h_digit] + " trăm")
    elif len(numbers) == 3: # Case like 0xx in a larger number
        res.append("không trăm")

    # Tens
    if t_digit == '0':
        if u_digit != '0' and (h_digit != '0' or len(numbers) == 3):
            res.append("lẻ")
    elif t_digit == '1':
        res.append("mười")
    else:
        res.append(units[t_digit] + " mươi")

    # Units
    if u_digit != '0':
        if u_digit == '1' and t_digit not in ('0', '1'):
            res.append("mốt")
        elif u_digit == '5' and (t_digit != '0' or (h_digit != '0' or len(numbers) == 3)):
            res.append("lăm")
        elif u_digit == '4' and t_digit not in ('0', '1'):
            # In some regions, 4 is 'tư' in certain positions,
            # but usually TTS uses 'bốn' unless specified.
            # The previous implementation used 'bốn'.
            res.append(units[u_digit])
        else:
            res.append(units[u_digit])

    return " ".join(res)

def n2w_large_number(numbers: str) -> str:
    """Convert large numbers to Vietnamese words."""
    if not numbers or not numbers.lstrip('0'):
        return units['0']
        
    # Remove leading zeros
    numbers = numbers.lstrip('0')

    # Split into 3-digit groups from right to left
    n_len = len(numbers)
    groups = []
    for i in range(n_len, 0, -3):
        groups.append(numbers[max(0, i-3):i])

    suffixes = ['', ' nghìn', ' triệu', ' tỷ']

    parts = []
    for i, group in enumerate(groups):
        if group == '000':
            continue

        word = n2w_hundreds(group)
        if word:
            # Suffix handling
            # 0: none, 1: nghìn, 2: triệu, 3: tỷ, 4: nghìn (tỷ), 5: triệu (tỷ), 6: tỷ (tỷ)...
            suffix_idx = i % 3
            main_suffix = suffixes[suffix_idx] if suffix_idx < len(suffixes) else ""

            # Billion (tỷ) is special in Vietnamese, it repeats or scales
            tỷ_count = i // 3

            if main_suffix or tỷ_count > 0:
                full_suffix = main_suffix + (" tỷ" * tỷ_count)
                word_with_suffix = f"{word}{full_suffix}"
            else:
                word_with_suffix = word
            parts.append(word_with_suffix)

    if not parts:
        return units['0']

    # Reverse parts and join
    parts.reverse()
    return ' '.join(parts).strip()

def n2w(number: str) -> str:
    """Main entry point for number to word conversion."""
    clean_number = pre_process_n2w(number)
    if not clean_number:
        return str(number)

    if len(clean_number) == 2 and clean_number[0] == '0':
        return f"không {units[clean_number[1]]}"

    return n2w_large_number(clean_number)

def n2w_single(number: str) -> str:
    """Convert number to word by digits (e.g. for phone numbers)."""
    if str(number).startswith('+84'):
        number = '0' + str(number)[3:]
    clean_number = pre_process_n2w(number)
    if not clean_number:
        return str(number)
    return process_n2w_single(clean_number)



import pytest
from unittest.mock import MagicMock, patch
import numpy as np
import torch
from vieneu.standard import VieNeuTTS
from pathlib import Path

@pytest.fixture
def mock_tts():
    # We use create=True because Llama might not be importable if llama-cpp-python is missing
    with patch('vieneu.standard.Llama', create=True) as mock_llama, \
         patch('vieneu.standard.DistillNeuCodec', create=True) as mock_codec, \
         patch('vieneu.standard.NeuCodec', create=True) as mock_neucodec, \
         patch('vieneu.base.hf_hub_download') as mock_hf_download:

        # Setup mock codec
        mock_codec_instance = MagicMock()
        mock_codec.from_pretrained.return_value = mock_codec_instance
        mock_codec_instance.hop_length = 480
        mock_codec_instance.sample_rate = 24000
        mock_codec_instance.decode_code.return_value = torch.zeros((1, 1, 1000))

        # Setup mock llama
        mock_llama_instance = MagicMock()
        mock_llama.from_pretrained.return_value = mock_llama_instance
        mock_llama_instance.return_value = {"choices": [{"text": "<|speech_1|><|speech_2|>"}]}

        # Setup mock voices.json
        mock_hf_download.return_value = "dummy_voices.json"

        with patch.object(VieNeuTTS, '_load_backbone'), \
             patch.object(VieNeuTTS, '_load_codec'), \
             patch.object(VieNeuTTS, '_warmup_model'), \
             patch('pathlib.Path.exists', return_value=True), \
             patch('builtins.open', MagicMock()), \
             patch('json.load', return_value={"presets": {"test_voice": {"codes": [1, 2], "text": "test"}}, "default_voice": "test_voice"}):
                tts = VieNeuTTS(backbone_repo="dummy-gguf", codec_repo="neuphonic/distill-neucodec")
                # Manually set some attributes that would have been set by _load_backbone/_load_codec
                tts.codec = mock_codec_instance
                tts.backbone = mock_llama_instance
                tts._is_quantized_model = True

                # Re-run _load_voices because it might have failed in __init__ due to patches not being active yet
                # Actually, the patches ARE active because they are in the 'with' block above.
                # But Path.exists(dummy_voices.json) might have failed if dummy_voices.json is not a real file.
                # Let's manually inject the voices
                tts._preset_voices = {"test_voice": {"codes": [1, 2], "text": "test"}}
                tts._default_voice = "test_voice"

                return tts

def test_tts_infer_mock(mock_tts):
    with patch.object(VieNeuTTS, '_decode', return_value=np.zeros(1000)):
        wav = mock_tts.infer("Xin chào", max_chars=10)
        assert isinstance(wav, np.ndarray)
        assert len(wav) > 0

def test_tts_infer_batch_mock(mock_tts):
    with patch.object(VieNeuTTS, '_decode', return_value=np.zeros(1000)):
        texts = ["Xin chào", "Chào buổi sáng"]
        results = mock_tts.infer_batch(texts)
        assert isinstance(results, list)
        assert len(results) == 2
        for res in results:
            assert isinstance(res, np.ndarray)

def test_tts_infer_multi_chunk_mock(mock_tts):
    with patch.object(VieNeuTTS, '_decode', return_value=np.zeros(1000)):
        # Force multiple chunks by using very small max_chars
        wav = mock_tts.infer("Xin chào Việt Nam thân yêu của tôi", max_chars=5)
        assert isinstance(wav, np.ndarray)
        assert len(wav) > 0

def test_tts_list_voices(mock_tts):
    voices = mock_tts.list_preset_voices()
    assert len(voices) > 0
    assert voices[0][1] == "test_voice"



import os
import json
import platform
import glob
import re
import logging
import functools
import sqlite3
import threading
from phonemizer import phonemize
from phonemizer.backend.espeak.espeak import EspeakWrapper
from vieneu_utils.normalize_text import VietnameseTTSNormalizer

# Configuration
DICT_DIR = os.getenv(
    'PHONEME_DICT_DIR',
    os.path.join(os.path.dirname(__file__), "phone_dict")
)

DB_PATH = os.path.join(DICT_DIR, "phone_dict.db")

# Configure logging
logger = logging.getLogger("Vieneu.Phonemizer")

class PhonemeDB:
    """SQLite-based dictionary for fast lookup and low memory usage."""
    def __init__(self, db_path: str):
        self.db_path = db_path
        self._local = threading.local()

    def _get_conn(self):
        if not hasattr(self._local, "conn"):
            self._local.conn = sqlite3.connect(self.db_path, check_same_thread=False)
        return self._local.conn

    def lookup_batch(self, words: list[str]) -> tuple[dict, dict]:
        """Fetch multiple words from DB in two logical groups: merged and common."""
        if not words: return {}, {}
        conn = self._get_conn()
        cursor = conn.cursor()
        
        merged_map = {}
        common_map = {}
        
        # SQLite has a limit on the number of host parameters (typically 999)
        chunk_size = 950
        for i in range(0, len(words), chunk_size):
            chunk = words[i : i + chunk_size]
            placeholders = ','.join(['?'] * len(chunk))

            # Query merged table
            cursor.execute(f"SELECT word, phone FROM merged WHERE word IN ({placeholders})", chunk)
            merged_map.update(dict(cursor.fetchall()))

            # Query common table
            cursor.execute(f"SELECT word, vi_phone, en_phone FROM common WHERE word IN ({placeholders})", chunk)
            for row in cursor.fetchall():
                common_map[row[0]] = {"vi": row[1], "en": row[2]}
        
        return merged_map, common_map

def setup_espeak_library() -> None:
    """Configure eSpeak library path based on operating system."""
    system = platform.system()
    
    if system == "Windows":
        _setup_windows_espeak()
    elif system == "Linux":
        _setup_linux_espeak()
    elif system == "Darwin":
        _setup_macos_espeak()
    else:
        logger.warning(f"Warning: Unsupported OS: {system}")
        return

def _setup_windows_espeak() -> None:
    """Setup eSpeak for Windows."""
    default_path = r"C:\Program Files\eSpeak NG\libespeak-ng.dll"
    if os.path.exists(default_path):
        EspeakWrapper.set_library(default_path)
    else:
        logger.warning("\033[91;1m⚠️ eSpeak-NG is not installed. The system will use the built-in dictionary, but it is recommended to install eSpeak-NG for maximum performance and accuracy.\033[0m")

def _setup_linux_espeak() -> None:
    """Setup eSpeak for Linux."""
    search_patterns = [
        "/usr/lib/x86_64-linux-gnu/libespeak-ng.so*",
        "/usr/lib/x86_64-linux-gnu/libespeak.so*",
        "/usr/lib/libespeak-ng.so*",
        "/usr/lib64/libespeak-ng.so*",
        "/usr/local/lib/libespeak-ng.so*",
    ]
    
    for pattern in search_patterns:
        matches = glob.glob(pattern)
        if matches:
            EspeakWrapper.set_library(sorted(matches, key=len)[0])
            return
    
    logger.warning("\033[91;1m⚠️ eSpeak-NG is not installed on Linux. The system will use the built-in dictionary, but it is recommended to install eSpeak-NG (sudo apt install espeak-ng) for maximum performance.\033[0m")

def _setup_macos_espeak() -> None:
    """Setup eSpeak for macOS."""
    espeak_lib = os.environ.get('PHONEMIZER_ESPEAK_LIBRARY')
    
    paths_to_check = [
        espeak_lib,
        "/opt/homebrew/lib/libespeak-ng.dylib",  # Apple Silicon
        "/usr/local/lib/libespeak-ng.dylib",     # Intel
        "/opt/local/lib/libespeak-ng.dylib",     # MacPorts
    ]
    
    for path in paths_to_check:
        if path and os.path.exists(path):
            EspeakWrapper.set_library(path)
            return
    
    logger.warning("\033[91;1m⚠️ eSpeak-NG is not installed on macOS. The system will use the built-in dictionary, but it is recommended to install eSpeak-NG (brew install espeak-ng) for maximum performance.\033[0m")

# Initialize
setup_espeak_library()
phone_db = PhonemeDB(DB_PATH)
normalizer = VietnameseTTSNormalizer()

def espeak_fallback_batch(texts: list[str], language: str = 'en-us') -> list[str]:
    """Batch fallback to espeak-ng for unknown segments."""
    if not texts: return []
    try:
        ph = phonemize(
            texts,
            language=language,
            backend='espeak',
            preserve_punctuation=True,
            with_stress=True,
            language_switch="remove-flags"
        )
        if isinstance(ph, str): ph = [ph]
        return [p.strip() for p in ph]
    except Exception as e:
        logger.warning(f"eSpeak fallback ({language}) failed: {e}")
        return texts

def propagate_language(tokens):
    """
    Propagate language labels for 'common' words based on the closest anchor.
    Sentence boundaries (strong punctuation) block propagation.
    """
    STOP_PUNCT = {'.', '!', '?', ';', ':', '(', ')', '[', ']', '{', '}'}
    
    # 1. Identify islands of common words
    islands = []
    current_island = []
    for i, token in enumerate(tokens):
        if token['lang'] == 'common':
            current_island.append(i)
        else:
            # Any non-common token (vi, en, or punct) breaks the island
            if current_island:
                islands.append(current_island)
                current_island = []
    if current_island:
        islands.append(current_island)

    # 2. For each island, find the closest valid anchor
    for island in islands:
        left_anchor, left_dist = None, 999
        right_anchor, right_dist = None, 999
        
        # Search left from the start of island
        for l in range(island[0] - 1, -1, -1):
            if tokens[l]['content'] in STOP_PUNCT: break
            if tokens[l]['lang'] in ('vi', 'en'):
                left_anchor = tokens[l]['lang']
                left_dist = island[0] - l
                break
        
        # Search right from the end of island
        for r in range(island[-1] + 1, len(tokens)):
            if tokens[r]['content'] in STOP_PUNCT: break
            if tokens[r]['lang'] in ('vi', 'en'):
                right_anchor = tokens[r]['lang']
                right_dist = r - island[-1]
                break
        
        # Decision logic: closest wins, tie-break to RIGHT for better switching
        final_lang = 'vi' # Default
        if left_anchor and right_anchor:
            # If distance is equal, we often prefer the language that follows (the target phrase)
            final_lang = right_anchor if right_dist <= left_dist else left_anchor
        elif left_anchor:
            final_lang = left_anchor
        elif right_anchor:
            final_lang = right_anchor
            
        for idx in island:
            tokens[idx]['lang'] = final_lang

@functools.lru_cache(maxsize=1024)
def _phonemize_with_dict_cached(text: str, skip_normalize: bool = False) -> str:
    return phonemize_batch([text], skip_normalize=skip_normalize, phoneme_dict=None)[0]

def phonemize_batch(texts: list[str], skip_normalize: bool = False, phoneme_dict: dict = None, **kwargs) -> list[str]:
    """Phonemize multiple texts with bilingual support and batch deduplication."""
    if not texts: return []
    if not skip_normalize: texts = [normalizer.normalize(t) for t in texts]

    use_system = (phoneme_dict is None)
    custom = phoneme_dict if phoneme_dict else {}

    batch_token_lists = []
    global_unknown = set()
    # Collect words that must bypass dict and go directly to espeak en-us
    force_espeak_words = set()

    for text in texts:
        matches = re.finditer(r'(<en>.*?</en>)|(\w+)|([^\w\s])', text, re.I | re.U)
        sent_tokens = []
        for m in matches:
            en_tag, word, punct = m.groups()
            if en_tag:
                content = re.sub(r'</?en>', '', en_tag, flags=re.I).strip()
                for st in re.finditer(r'(\w+)|([^\w\s])', content, re.U):
                    sw, sp = st.groups()
                    if sp:
                        sent_tokens.append({'lang': 'punct', 'content': sp, 'phone': sp})
                    else:
                        # Mark token as force_espeak: bỏ qua dict, dùng espeak en-us trực tiếp
                        sent_tokens.append({'lang': 'en', 'content': sw, 'phone': None, 'force_espeak': True})
                        force_espeak_words.add(sw)
            elif punct:
                sent_tokens.append({'lang': 'punct', 'content': punct, 'phone': punct})
            elif word:
                sent_tokens.append({'lang': 'unknown', 'content': word, 'phone': None})
        batch_token_lists.append(sent_tokens)

    # Resolve all NON-forced words from DB in one batch
    all_words = set()
    for sent in batch_token_lists:
        for t in sent:
            if t['lang'] != 'punct' and not t.get('force_espeak'):
                all_words.add(t['content'].lower())

    db_merged, db_common = phone_db.lookup_batch(list(all_words)) if use_system else ({}, {})

    # Fill tokens with data from DB or custom dict (skipping force_espeak tokens)
    for sent in batch_token_lists:
        for t in sent:
            if t['lang'] == 'punct': continue
            # force_espeak tokens skip all dict lookups entirely
            if t.get('force_espeak'): continue
            lw = t['content'].lower()

            # Priority: 1. Custom dict
            if lw in custom:
                t['phone'] = custom[lw]
                t['lang'] = 'en'
            # 2. Merged DB
            elif lw in db_merged:
                val = db_merged[lw]
                t['phone'] = val
                t['lang'] = 'en' if val.startswith('<en>') else 'vi'
            # 3. Common DB
            elif lw in db_common:
                t['phone'] = db_common[lw]
                t['lang'] = 'common'
            # 4. Global Unknown
            else:
                global_unknown.add(t['content'])
                t['lang'] = 'en'  # Temporary placeholder for espeak fallback logic

    # --- Batch espeak for force_espeak words (always en-us) ---
    if force_espeak_words:
        fe_words = sorted(list(force_espeak_words))
        fe_phones = espeak_fallback_batch(fe_words, 'en-us')
        fe_lut = {w: f"<en>{p}" for w, p in zip(fe_words, fe_phones)}
        for sent in batch_token_lists:
            for t in sent:
                if t.get('force_espeak') and t['phone'] is None:
                    t['phone'] = fe_lut.get(t['content'], t['content'])

    # --- Batch espeak for global_unknown words (vi or en-us by accent) ---
    if global_unknown:
        u_words = sorted(list(global_unknown))
        vi_accents = "àáảãạăằắẳẵặâầấẩẫậèéẻẽẹêềếểễệìíỉĩịòóỏõọôồốổỗộơờớởỡợùúủũụưừứửữựỳýỷỹỵđ"
        def has_accent(w): return any(c in vi_accents for c in w.lower())

        vi_words = [w for w in u_words if has_accent(w)]
        en_words = [w for w in u_words if not has_accent(w)]

        lut = {}
        if vi_words:
            res_vi = espeak_fallback_batch(vi_words, 'vi')
            for w, p in zip(vi_words, res_vi):
                lut[w] = p
        if en_words:
            res_en = espeak_fallback_batch(en_words, 'en-us')
            for w, p in zip(en_words, res_en):
                lut[w] = f"<en>{p}"

        for sent in batch_token_lists:
            for t in sent:
                if t['phone'] is None and t['content'] in lut:
                    t['phone'] = lut[t['content']]
                    if has_accent(t['content']):
                        t['lang'] = 'vi'
                    else:
                        t['lang'] = 'en'

    results = []
    for sent in batch_token_lists:
        propagate_language(sent)
        sent_phones = []
        for t in sent:
            if t['lang'] == 'punct':
                sent_phones.append(t['phone'])
            else:
                p = t['phone']
                if isinstance(p, dict):
                    p = p['en'] if t['lang'] == 'en' else p['vi']
                if p is None: p = t['content']
                p = p.replace('<en>', '')
                sent_phones.append(p)
        txt = " ".join(sent_phones)
        txt = re.sub(r'\s+([.,!?;:])', r'\1', txt)
        results.append(txt.strip())
    return results

def phonemize_text(text: str) -> str:
    return phonemize_batch([text])[0]

def phonemize_with_dict(text: str, phoneme_dict: dict = None, skip_normalize: bool = False) -> str:
    if phoneme_dict is not None:
        return phonemize_batch([text], skip_normalize=skip_normalize, phoneme_dict=phoneme_dict)[0]
    return _phonemize_with_dict_cached(text, skip_normalize=skip_normalize)

if __name__ == "__main__":
    import sys
    test_text = " ".join(sys.argv[1:]) if len(sys.argv) > 1 else "Một nghiên cứu gần đây của Đại học Harvard phối hợp với Boston Consulting Group (BCG) thực hiện trên 758 tư vấn viên cao cấp đã chỉ ra mặt trái tàn khốc này. Kết quả cho thấy: những người sử dụng AI để giải quyết các vấn đề kinh doanh phức tạp (vượt quá khả năng của AI) mắc sai lầm nhiều hơn 19% so với những người không dùng. Điều đáng nói là, vì câu trả lời của AI quá trôi chảy và tự tin, những tư vấn viên này đã hoàn toàn tin tưởng và bỏ qua bước kiểm chứng. Họ để cho sự tiện lợi đánh lừa chuyên môn của mình."
    print(f"Output: {phonemize_text(test_text)}")


import pytest
from unittest.mock import MagicMock, patch
import numpy as np
import torch
from vieneu.factory import Vieneu
from vieneu.standard import VieNeuTTS
from vieneu.remote import RemoteVieNeuTTS
from vieneu_utils.phonemize_text import PhonemeDB
import sqlite3

@pytest.fixture
def mock_torch_backbone():
    with patch("transformers.AutoTokenizer.from_pretrained") as mock_tok, \
         patch("transformers.AutoModelForCausalLM.from_pretrained") as mock_model, \
         patch("neucodec.NeuCodec.from_pretrained") as mock_codec, \
         patch("neucodec.DistillNeuCodec.from_pretrained") as mock_distill_codec:

        tokenizer = MagicMock()
        tokenizer.pad.return_value = {"input_ids": torch.zeros((2, 10), dtype=torch.long)}
        tokenizer.convert_tokens_to_ids.return_value = 100
        tokenizer.decode.return_value = "<|speech_1|><|speech_2|>"
        mock_tok.return_value = tokenizer

        model = MagicMock()
        model.device = torch.device("cpu")
        model.to.return_value = model
        model.generate.return_value = torch.zeros((2, 20), dtype=torch.long)
        mock_model.return_value = model

        codec = MagicMock()
        codec.device = torch.device("cpu")
        codec.sample_rate = 24000
        codec.decode_code.return_value = torch.zeros((1, 1, 4800))
        mock_codec.return_value = codec
        mock_distill_codec.return_value = codec

        yield {"tokenizer": tokenizer, "model": model, "codec": codec}

def test_sqlite_chunking():
    """Test that PhonemeDB.lookup_batch handles more than 999 words by chunking."""
    db = PhonemeDB(":memory:")
    conn = sqlite3.connect(":memory:")
    # Mock the connection to return a real-ish cursor
    db._get_conn = MagicMock(return_value=conn)

    conn.execute("CREATE TABLE merged (word TEXT, phone TEXT)")
    conn.execute("CREATE TABLE common (word TEXT, vi_phone TEXT, en_phone TEXT)")

    # Insert 1000 words
    words = [f"word_{i}" for i in range(1500)]
    for w in words[:1000]:
        conn.execute("INSERT INTO merged (word, phone) VALUES (?, ?)", (w, f"phone_{w}"))

    # We should be able to look up all 1500 words without error
    merged, common = db.lookup_batch(words)
    assert len(merged) == 1000
    assert "word_0" in merged
    assert "word_999" in merged

def test_base_encode_reference_device(mock_torch_backbone):
    """Test BaseVieneuTTS.encode_reference moves tensor to correct device."""
    from vieneu.standard import VieNeuTTS

    with patch("librosa.load", return_value=(np.zeros(16000), 16000)), \
         patch.object(VieNeuTTS, '_warmup_model'):
        tts = VieNeuTTS(backbone_repo="dummy", backbone_device="cpu")
        tts.codec.device = torch.device("cpu")

        with patch.object(torch.Tensor, 'to', wraps=torch.zeros(1).to) as mock_to:
            tts.encode_reference("dummy.wav")
            # Verify .to() was called to move the tensor to codec device
            mock_to.assert_called()

def test_standard_true_batching(mock_torch_backbone):
    """Test that VieNeuTTS.infer_batch uses true batch generation for Torch backend."""
    from vieneu.standard import VieNeuTTS

    with patch.object(VieNeuTTS, '_warmup_model'), \
         patch("vieneu.standard.phonemize_batch", return_value=["p1", "p2"]):
        tts = VieNeuTTS(backbone_repo="dummy", backbone_device="cpu")
        tts._is_quantized_model = False # Force torch backend

        texts = ["Text 1", "Text 2"]
        with patch.object(tts, '_decode', return_value=np.zeros(1000)):
            results = tts.infer_batch(texts, ref_codes=[1], ref_text="ref")

            assert len(results) == 2

def test_fast_infer_batch_phonemize_called(mock_torch_backbone):
    """Test that FastVieNeuTTS.infer_batch calls phonemize_batch."""
    from vieneu.fast import FastVieNeuTTS

    with patch("lmdeploy.pipeline") as mock_pipeline, \
         patch("lmdeploy.GenerationConfig"), \
         patch.object(FastVieNeuTTS, '_warmup_model'):

        mock_pipeline_instance = MagicMock()
        mock_pipeline.return_value = mock_pipeline_instance
        mock_pipeline_instance.return_value = [MagicMock(text="codes")]

        tts = FastVieNeuTTS(backbone_device="cuda")

        texts = ["Text 1", "Text 2"]
        with patch("vieneu.fast.phonemize_batch", return_value=["p1", "p2"]) as mock_ph_batch, \
             patch.object(tts, '_decode', return_value=np.zeros(1000)):
            tts.infer_batch(texts, ref_codes=[1], ref_text="ref")
            mock_ph_batch.assert_called_once()
            # Verify LMDeploy pipeline was called
            mock_pipeline_instance.assert_called()

def test_remote_parallel_chunking():
    """Test that RemoteVieNeuTTS.infer uses parallel async processing for multi-chunks."""
    with patch("neucodec.DistillNeuCodec.from_pretrained"):
        tts = RemoteVieNeuTTS(api_base="http://mock", model_name="mock")

        # Force 2 chunks
        with patch("vieneu.remote.split_text_into_chunks", return_value=["chunk1", "chunk2"]), \
             patch.object(RemoteVieNeuTTS, 'infer_async', return_value=np.zeros(2000)) as mock_async, \
             patch.object(RemoteVieNeuTTS, '_resolve_ref_voice', return_value=(torch.zeros(10), "ref")):

            res = tts.infer("Long text that splits")
            assert mock_async.called
            # Should be called once via asyncio.run(self.infer_async(...))
            # which then parallelizes internal chunks
            mock_async.assert_called_once()

def test_lora_loading_logic(mock_torch_backbone):
    """Test LoRA adapter loading and unloading."""
    from vieneu.standard import VieNeuTTS

    with patch.object(VieNeuTTS, '_warmup_model'), \
         patch("peft.PeftModel.from_pretrained") as mock_peft:
        tts = VieNeuTTS(backbone_repo="dummy", backbone_device="cpu")

        # Load LoRA
        tts.load_lora_adapter("lora_repo")
        assert tts._lora_loaded is True
        mock_peft.assert_called_once()

        # Unload LoRA
        with patch.object(tts.backbone, 'unload', return_value=mock_torch_backbone["model"]):
            tts.unload_lora_adapter()
            assert tts._lora_loaded is False



import pytest
from unittest.mock import MagicMock, patch
import numpy as np
import torch
from vieneu.standard import VieNeuTTS
from vieneu.remote import RemoteVieNeuTTS

@pytest.fixture
def mock_codec():
    codec = MagicMock()
    codec.sample_rate = 24000
    codec.device = "cpu"
    # Mock encode_code to return a tensor of codes
    codec.encode_code.return_value = torch.zeros((1, 1, 10), dtype=torch.long)
    # Mock decode_code to return a tensor of audio
    codec.decode_code.return_value = torch.zeros((1, 1, 4800), dtype=torch.float32)
    return codec

@pytest.fixture
def mock_backbone():
    backbone = MagicMock()
    backbone.device = torch.device("cpu")
    backbone.to.return_value = backbone
    # Mock generate for torch backend
    backbone.generate.return_value = torch.tensor([[0, 1, 2, 3]]) # Dummy output tokens
    # Mock llama-cpp call
    backbone.return_value = {"choices": [{"text": "<|speech_1|><|speech_2|>"}]}
    return backbone

@pytest.fixture
def mock_tokenizer():
    tokenizer = MagicMock()

    # Map special tokens to specific IDs
    token_to_id = {
        "<|SPEECH_REPLACE|>": 1001,
        "<|SPEECH_GENERATION_START|>": 1002,
        "<|TEXT_REPLACE|>": 1003,
        "<|TEXT_PROMPT_START|>": 1004,
        "<|TEXT_PROMPT_END|>": 1005,
        "<|SPEECH_GENERATION_END|>": 1006
    }
    tokenizer.convert_tokens_to_ids.side_effect = lambda x: token_to_id.get(x, 999)

    # Mock encode to include the TEXT_REPLACE and SPEECH_REPLACE tokens when expected
    def mocked_encode(text, **kwargs):
        if "TEXT_REPLACE" in text:
            return [10, 1003, 11, 1001]
        return [1, 2, 3]

    tokenizer.encode.side_effect = mocked_encode
    tokenizer.decode.return_value = "<|speech_1|><|speech_2|>"
    return tokenizer

def test_vieneu_tts_init(mock_codec, mock_backbone, mock_tokenizer):
    with patch("vieneu.standard.NeuCodec.from_pretrained", return_value=mock_codec), \
         patch("vieneu.standard.DistillNeuCodec.from_pretrained", return_value=mock_codec), \
         patch("transformers.AutoTokenizer.from_pretrained", return_value=mock_tokenizer), \
         patch("transformers.AutoModelForCausalLM.from_pretrained", return_value=mock_backbone):

        tts = VieNeuTTS(backbone_repo="some/repo", backbone_device="cpu")
        assert tts.backbone is not None
        assert tts.codec is not None

def test_vieneu_tts_infer(mock_codec, mock_backbone, mock_tokenizer):
    with patch("vieneu.standard.NeuCodec.from_pretrained", return_value=mock_codec), \
         patch("vieneu.standard.DistillNeuCodec.from_pretrained", return_value=mock_codec), \
         patch("transformers.AutoTokenizer.from_pretrained", return_value=mock_tokenizer), \
         patch("transformers.AutoModelForCausalLM.from_pretrained", return_value=mock_backbone):

        tts = VieNeuTTS(backbone_repo="some/repo", backbone_device="cpu")

        with patch("vieneu.standard.phonemize_with_dict", return_value="phonemes"):
            audio = tts.infer("Xin chào", ref_codes=[1, 2, 3], ref_text="Chào")
            assert isinstance(audio, np.ndarray)
            assert len(audio) > 0
            assert len(audio) == 4800

def test_vieneu_tts_infer_with_voice_preset(mock_codec, mock_backbone, mock_tokenizer):
    with patch("vieneu.standard.NeuCodec.from_pretrained", return_value=mock_codec), \
         patch("vieneu.standard.DistillNeuCodec.from_pretrained", return_value=mock_codec), \
         patch("transformers.AutoTokenizer.from_pretrained", return_value=mock_tokenizer), \
         patch("transformers.AutoModelForCausalLM.from_pretrained", return_value=mock_backbone):

        tts = VieNeuTTS(backbone_repo="some/repo", backbone_device="cpu")
        tts._preset_voices = {"test_voice": {"codes": [1, 2, 3], "text": "test"}}

        with patch("vieneu.standard.phonemize_with_dict", return_value="phonemes"):
            audio = tts.infer("Xin chào", voice=tts.get_preset_voice("test_voice"))
            assert isinstance(audio, np.ndarray)
            assert len(audio) == 4800

def test_remote_vieneu_tts_infer(mock_codec):
    with patch("vieneu.standard.DistillNeuCodec.from_pretrained", return_value=mock_codec):
        tts = RemoteVieNeuTTS(api_base="http://mock-api", model_name="mock-model")

        mock_response = MagicMock()
        mock_response.json.return_value = {
            "choices": [{"message": {"content": "<|speech_1|><|speech_2|>"}}]
        }
        mock_response.raise_for_status = MagicMock()

        with patch("requests.post", return_value=mock_response), \
             patch("vieneu.remote.phonemize_with_dict", return_value="phonemes"):
            audio = tts.infer("Xin chào", ref_codes=[1, 2, 3], ref_text="Chào")
            assert isinstance(audio, np.ndarray)
            assert len(audio) == 4800

def test_vieneu_tts_streaming(mock_codec, mock_backbone, mock_tokenizer):
    with patch("vieneu.standard.NeuCodec.from_pretrained", return_value=mock_codec), \
         patch("vieneu.standard.DistillNeuCodec.from_pretrained", return_value=mock_codec), \
         patch("transformers.AutoTokenizer.from_pretrained", return_value=mock_tokenizer), \
         patch("transformers.AutoModelForCausalLM.from_pretrained", return_value=mock_backbone):

        tts = VieNeuTTS(backbone_repo="some/repo", backbone_device="cpu")

        with patch("vieneu.standard.phonemize_with_dict", return_value="phonemes"):
            stream = tts.infer_stream("Xin chào", ref_codes=[1, 2, 3], ref_text="Chào")
            chunks = list(stream)
            assert len(chunks) > 0
            assert isinstance(chunks[0], np.ndarray)



import numpy as np
import pytest
from vieneu.utils import _linear_overlap_add
from vieneu_utils.core_utils import join_audio_chunks

def test_linear_overlap_add():
    # Create two overlapping frames
    frame_len = 100
    stride = 50
    frame1 = np.ones(frame_len, dtype=np.float32)
    frame2 = np.ones(frame_len, dtype=np.float32)

    frames = [frame1, frame2]
    out = _linear_overlap_add(frames, stride)

    # Total length should be stride * (len(frames) - 1) + frame_len = 50 * 1 + 100 = 150
    assert out.shape == (150,)

    # Check that it's not all zeros
    assert np.any(out != 0)

    # With all ones and linear OLA, the result should be close to 1.0 where it overlaps
    # (weight1 * 1 + weight2 * 1) / (weight1 + weight2) = 1.0
    assert np.allclose(out[50:100], 1.0)

def test_linear_overlap_add_empty():
    assert _linear_overlap_add([], 50).shape == (0,)

def test_join_audio_chunks_simple():
    chunks = [np.ones(100), np.zeros(100)]
    joined = join_audio_chunks(chunks, sr=16000)
    assert joined.shape == (200,)
    assert np.array_equal(joined[:100], np.ones(100))
    assert np.array_equal(joined[100:], np.zeros(100))

def test_join_audio_chunks_silence():
    chunks = [np.ones(100), np.ones(100)]
    sr = 16000
    silence_p = 0.1 # 0.1s * 16000 = 1600 samples
    joined = join_audio_chunks(chunks, sr=sr, silence_p=silence_p)
    assert joined.shape == (100 + 1600 + 100,)
    assert np.all(joined[100:1700] == 0)

def test_join_audio_chunks_crossfade():
    chunks = [np.ones(1000), np.zeros(1000)]
    sr = 16000
    crossfade_p = 0.01 # 0.01s * 16000 = 160 samples
    joined = join_audio_chunks(chunks, sr=sr, crossfade_p=crossfade_p)
    # Length should be 1000 + 1000 - 160 = 1840
    assert joined.shape == (1840,)
    # Check transition
    assert joined[0] == 1.0
    assert joined[-1] == 0.0
    # Mid-point of crossfade should be 0.5
    assert np.allclose(joined[1000 - 80], 0.5, atol=0.01)

def test_join_audio_chunks_empty():
    assert join_audio_chunks([], 16000).shape == (0,)

def test_join_audio_chunks_single():
    chunk = np.ones(100)
    assert np.array_equal(join_audio_chunks([chunk], 16000), chunk)



import re
from .num2vi import n2w, n2w_single

from .numerical import normalize_number_vi
from .datestime import normalize_date, normalize_time
from .text_norm import normalize_others, expand_measurement, expand_currency, expand_compound_units, expand_abbreviations, expand_standalone_letters

def clean_vietnamese_text(text):
    mask_map = {}
    import string
    
    def protect(match):
        idx = len(mask_map)
        mask = "mask" + "".join([string.ascii_lowercase[int(d)] for d in str(idx).zfill(4)]) + "mask"
        mask_map[mask] = match.group(0)
        return mask
    
    text = re.sub(r'___PROTECTED_EN_TAG_\d+___', protect, text)
    
    # Handle common abbreviations early to avoid unit conflicts
    text = expand_abbreviations(text)
    
    text = normalize_date(text)
    text = normalize_time(text)

    text = re.sub(r'(\d+(?:,\d+)?)\s*[–\-—~]\s*(\d+(?:,\d+)?)', r'\1 đến \2', text)
    
    # 3. Replace standalone hyphens with commas (for better TTS prosody/pausing)
    text = re.sub(r'(?<=\s)[–\-—](?=\s)', ',', text)
    
    text = re.sub(r'\s*(?:->|=>)\s*', ' sang ', text)

    # Expand measurements and currencies BEFORE general floats
    text = expand_compound_units(text)
    text = expand_measurement(text)
    text = expand_currency(text)

    def _expand_float(m):
        int_part = n2w(m.group(1))
        dec_part = n2w(m.group(2))
        res = f"{int_part} phẩy {dec_part}"
        if m.group(3):
            res += " phần trăm"
        return f" {res} "
    text = re.sub(r'\b(\d+),(\d+)(%)?', _expand_float, text)
    
    def _strip_dot_sep(m):
        return m.group(0).replace('.', '')
    text = re.sub(r'\b\d+(?:\.\d{3})+\b', _strip_dot_sep, text)
    
    text = normalize_others(text)
    text = normalize_number_vi(text)
    
    # Protect internally generated <en> tags from expand_standalone_letters
    text = re.sub(r'<en>.*?</en>', protect, text, flags=re.IGNORECASE)
    
    # Finally expand standalone letters to catch initials like "M."
    text = expand_standalone_letters(text)

    # Collapse redundant punctuation and whitespace
    # 1. Collapse multiple spaces BUT preserve newlines
    text = re.sub(r'[ \t\xA0]+', ' ', text)
    # 2. Collapse consecutive commas and handle comma-punctuation pairs
    text = re.sub(r',\s*,', ',', text)
    text = re.sub(r',\s*([.!?;])', r'\1', text)
    # 3. Handle redundant spaces before punctuation
    text = re.sub(r'\s+([,.!?;:])', r'\1', text)
    # 4. Remove leading/trailing commas if they end up at the start/end of sentence parts
    text = text.strip().strip(',')

    for mask, original in mask_map.items():
        text = text.replace(mask, original)
        text = text.replace(mask.lower(), original)
        
    return text



import re
from .num2vi import n2w, n2w_single
from .symbols import vietnamese_re, vietnamese_without_num_re

_en_letter_names = {
    "a": "ây", "b": "bi", "c": "xi", "d": "đi", "e": "i", "f": "ép", "g": "di",
    "h": "ếch", "i": "ai", "j": "giây", "k": "cây", "l": "eo", "m": "em", "n": "en",
    "o": "âu", "p": "pi", "q": "kiu", "r": "a", "s": "ét", "t": "ti", "u": "du",
    "v": "vi", "w": "đắp bờ liu", "x": "ích", "y": "quai", "z": "di",
    "0": "zero", "1": "one", "2": "two", "3": "three", "4": "four",
    "5": "five", "6": "six", "7": "seven", "8": "eight", "9": "nine"
}

_vi_letter_names = {
    "a": "a", "b": "bê", "c": "xê", "d": "dê", "đ": "đê", "e": "e", "ê": "ê",
    "f": "ép", "g": "gờ", "h": "hát", "i": "i", "j": "giây", "k": "ca", "l": "lờ",
    "m": "mờ", "n": "nờ", "o": "ô", "ô": "ô", "ơ": "ơ", "p": "phê", "q": "qui",
    "r": "rờ", "s": "ét", "t": "tê", "u": "u", "ư": "ư", "v": "vờ", "w": "vê kép",
    "x": "ích", "y": "y", "z": "dét"
}

_common_email_domains = {
    "gmail.com": "gờ meo chấm com",
    "yahoo.com": "da hu chấm com",
    "yahoo.com.vn": "da hu chấm com chấm vê nờ",
    "outlook.com": "aut lúc chấm com",
    "hotmail.com": "hót meo chấm com",
    "icloud.com": "ai clao chấm com",
    "fpt.vn": "ép phê tê chấm vê nờ",
    "fpt.com.vn": "ép phê tê chấm com chấm vê nờ",
}

_measurement_key_vi = {
    "km": "ki lô mét", "dm": "đê xi mét", "cm": "xen ti mét", "mm": "mi li mét",
    "nm": "na nô mét", "µm": "mic rô mét", "μm": "mic rô mét", "m": "mét",
    "kg": "ki lô gam", "g": "gam", "mg": "mi li gam",
    "km2": "ki lô mét vuông", "m2": "mét vuông", "cm2": "xen ti mét vuông", "mm2": "mi li mét vuông",
    "ha": "héc ta",
    "km3": "ki lô mét khối", "m3": "mét khối", "cm3": "xen ti mét khối", "mm3": "mi li mét khối",
    "l": "lít", "dl": "đê xi lít", "ml": "mi li lít", "hl": "héc tô lít",
    "kw": "ki lô oát", "mw": "mê ga oát", "gw": "gi ga oát",
    "kwh": "ki lô oát giờ", "mwh": "mê ga oát giờ", "wh": "oát giờ",
    "hz": "héc", "khz": "ki lô héc", "mhz": "mê ga héc", "ghz": "gi ga héc",
    "pa": "pát cal", "kpa": "ki lô pát cal", "mpa": "mê ga pát cal",
    "bar": "ba", "mbar": "mi li ba", "atm": "át mốt phia", "psi": "pi ét xai",
    "j": "giun", "kj": "ki lô giun",
    "cal": "ca lo", "kcal": "ki lô ca lo",
    "h": "giờ", "p": "phút", "s": "giây",
    "sqm": "mét vuông", "cum": "mét khối",
    "gb": "gi ga bai", "mb": "mê ga bai", "kb": "ki lô bai", "tb": "tê ra bai",
    "db": "đê xi ben", "oz": "ao xơ", "lb": "pao", "lbs": "pao",
    "ft": "phít", "in": "ins", "dpi": "đê phê i"
}

_currency_key = {
    "usd": "đô la Mỹ", "vnd": "đồng", "đ": "đồng", "euro": "ơ rô", "%": "phần trăm"
}

_letter_key_vi = _vi_letter_names

_acronyms_exceptions_vi = {
    "CĐV": "cổ động viên", "HĐND": "hội đồng nhân dân", "TAND": "toàn án nhân dân",
    "BHXH": "bảo hiểm xã hội", "BHTN": "bảo hiểm thất nghiệp", "TP.HCM": "thành phố hồ chí minh",
    "VN": "việt nam", "UBND": "uỷ ban nhân dân", "TP": "thành phố", "HCM": "hồ chí minh",
    "HN": "hà nội", "BTC": "ban tổ chức", "CLB": "câu lạc bộ", "HTX": "hợp tác xã",
    "NXB": "nhà xuất bản", "TW": "trung ương", "CSGT": "cảnh sát giao thông", "LHQ": "liên hợp quốc",
    "THCS": "trung học cơ sở", "THPT": "trung học phổ thông", "ĐH": "đại học", "HLV": "huấn luyện viên",
    "GS": "giáo sư", "TS": "tiến sĩ", "TNHH": "trách nhiệm hữu hạn", "VĐV": "vận động viên",
    "TPHCM": "thành phố hồ chí minh", "PGS": "phó giáo sư"
}

# Compiled Regular Expressions
RE_ROMAN_NUMBER = re.compile(r"\b(?=[IVXLCDM]{2,})M{0,4}(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})\b")
RE_LETTER = re.compile(r"(chữ|chữ cái|kí tự|ký tự)\s+(['\"]?)([a-z])(['\"]?)\b", re.IGNORECASE)
RE_STANDALONE_LETTER = re.compile(r'\b([a-zA-Z])\b\.?')
RE_URL = re.compile(r'\b(?:https?://|www\.)[A-Za-z0-9.\-_~:/?#\[\]@!$&\'()*+,;=]+\b')
RE_SLASH_NUMBER = re.compile(r'\b(\d+)/(\d+)\b')
RE_EMAIL = re.compile(r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b')
RE_SENTENCE_SPLIT = re.compile(r'([.!?]+(?:\s+|$))')
RE_ACRONYM = re.compile(r'\b(?=[A-Z0-9]*[A-Z])[A-Z0-9]{2,}\b')
RE_ALPHANUMERIC = re.compile(r'\b(\d+)([a-zA-Z])\b')
RE_BRACKETS = re.compile(r'[\(\[\{]\s*(.*?)\s*[\)\]\}]')
RE_STRIP_BRACKETS = re.compile(r'[\[\]\(\)\{\}]')
RE_TEMP_C_NEG = re.compile(r'-(\d+(?:[.,]\d+)?)\s*°\s*c\b', re.IGNORECASE)
RE_TEMP_F_NEG = re.compile(r'-(\d+(?:[.,]\d+)?)\s*°\s*f\b', re.IGNORECASE)
RE_TEMP_C = re.compile(r'(\d+(?:[.,]\d+)?)\s*°\s*c\b', re.IGNORECASE)
RE_TEMP_F = re.compile(r'(\d+(?:[.,]\d+)?)\s*°\s*f\b', re.IGNORECASE)
RE_DEGREE = re.compile(r'°')
RE_VERSION = re.compile(r'\b(\d+(?:\.\d+)+)\b')
RE_CLEAN_OTHERS = re.compile(r'[^\w\sàáảãạăắằẳẵặâấầẩẫậèéẻẽẹêếềểễệìíỉĩịòóỏõọôốồổỗộơớờởỡợùúủũụưứừửữỳýỷỹỵđ.,!?;:@%_]')

# Reusable patterns for measurement/currency
_MAGNITUDE_P = r"\s*(tỷ|triệu|nghìn|ngàn)?\s*"
_NUMERIC_P = r"((?:\d+[.,])*\d+)"

# Pre-compiled regex for compound units
RE_COMPOUND_UNIT = re.compile(rf"{_NUMERIC_P}?\s*\b([a-zμµ²³°]+)/([a-zμµ²³°0-9]+)\b", re.IGNORECASE)

# Pre-compiled currency patterns
RE_CURRENCY_PREFIX_USD = re.compile(rf"\$\s*{_NUMERIC_P}{_MAGNITUDE_P}", re.IGNORECASE)
RE_CURRENCY_SUFFIX_USD = re.compile(rf"{_NUMERIC_P}{_MAGNITUDE_P}\$", re.IGNORECASE)
RE_PERCENTAGE = re.compile(rf"{_NUMERIC_P}\s*%", re.IGNORECASE)

# Pre-compile measurement and currency unit patterns
_MEASUREMENT_PATTERNS = []
for unit, full in sorted(_measurement_key_vi.items(), key=lambda x: len(x[0]), reverse=True):
    pattern = re.compile(rf"\b{_NUMERIC_P}{_MAGNITUDE_P}{unit}\b", re.IGNORECASE)

    standalone_pattern = None
    safe_standalone = [
        "km", "cm", "mm", "kg", "mg",
        "m2", "km2", "usd", "vnd",
        "mhz", "khz", "ghz", "hz"
    ]
    if unit.lower() in safe_standalone:
        standalone_pattern = re.compile(rf"(?<![\d.,])\b{unit}\b", re.IGNORECASE)

    _MEASUREMENT_PATTERNS.append((pattern, standalone_pattern, full))

_CURRENCY_PATTERNS = []
for unit, full in _currency_key.items():
    if unit == "%": continue
    pattern = re.compile(rf"\b{_NUMERIC_P}{_MAGNITUDE_P}{unit}\b", re.IGNORECASE)
    _CURRENCY_PATTERNS.append((pattern, full))

# Pre-compile acronyms exceptions
_ACRONYMS_EXCEPTIONS_RE = [(re.compile(rf"\b{re.escape(k)}\b"), v) for k, v in _acronyms_exceptions_vi.items()]

def _expand_number_with_sep(num_str):
    if not num_str: return ""
    if "," in num_str:
        # Standard Vietnamese float: 1,5 or 1.000,5
        clean_num = num_str.replace(".", "")
        parts = clean_num.split(",")
        if len(parts) == 2:
            return f"{n2w(parts[0])} phẩy {n2w(parts[1])}"
    
    if "." in num_str:
        # Check if it's a thousand separator format (e.g. 1.000, 1.000.000)
        # Vietnamese thousand sep is ALWAYS exactly 3 digits after the dot.
        if re.fullmatch(r"\d+(?:\.\d{3})+", num_str):
            return n2w(num_str.replace(".", ""))
        # Otherwise treat dot as "chấm" (e.g. version 1.3 or English-style decimal 1.5)
        return " chấm ".join([n2w(p) for p in num_str.split(".")])
        
    return n2w(num_str)

def expand_measurement(text):
    def _repl(m, full):
        num = m.group(1)
        mag = m.group(2) if m.group(2) else ""
        expanded_num = _expand_number_with_sep(num)
        return f"{expanded_num} {mag} {full}".replace("  ", " ").strip()
    
    for pattern, standalone_pattern, full in _MEASUREMENT_PATTERNS:
        # Case with number
        text = pattern.sub(lambda m, f=full: _repl(m, f), text)
        
        # Standalone units
        if standalone_pattern:
            text = standalone_pattern.sub(f" {full} ", text)
    return text

def expand_currency(text):
    def _repl(m, full):
        num = m.group(1)
        mag = m.group(2) if m.group(2) else ""
        expanded_num = _expand_number_with_sep(num)
        return f"{expanded_num} {mag} {full}".replace("  ", " ").strip()
        
    text = RE_CURRENCY_PREFIX_USD.sub(lambda m: _repl(m, "đô la Mỹ"), text)
    text = RE_CURRENCY_SUFFIX_USD.sub(lambda m: _repl(m, "đô la Mỹ"), text)
    text = RE_PERCENTAGE.sub(lambda m: f"{_expand_number_with_sep(m.group(1))} phần trăm", text)
    
    for pattern, full in _CURRENCY_PATTERNS:
        text = pattern.sub(lambda m, f=full: _repl(m, f), text)
    return text

def expand_compound_units(text):
    def _repl_compound(m):
        num_str = m.group(1) if m.group(1) else ""
        num = _expand_number_with_sep(num_str)
        u1 = m.group(2).lower()
        u2 = m.group(3).lower()
        full1 = _measurement_key_vi.get(u1, u1)
        full2 = _measurement_key_vi.get(u2, u2)
        res = f" {full1} trên {full2} "
        if num:
            res = f"{num} {res}"
        return res

    text = RE_COMPOUND_UNIT.sub(_repl_compound, text)
    return text

def expand_roman(match):
    roman_numerals = {"I": 1, "V": 5, "X": 10, "L": 50, "C": 100, "D": 500, "M": 1000}
    num = match.group(0).upper()
    if not num: return ""
    result = 0
    for i, c in enumerate(num):
        if (i + 1) == len(num) or roman_numerals[c] >= roman_numerals[num[i + 1]]:
            result += roman_numerals[c]
        else:
            result -= roman_numerals[c]
    return f" {n2w(str(result))} "

def expand_letter(match):
    prefix, q1, char, q2 = match.groups()
    if char.lower() in _letter_key_vi:
        return f"{prefix} {_letter_key_vi[char.lower()]} "
    return match.group(0)

def expand_abbreviations(text):
    abbrs = {"v.v": " vân vân", "v/v": " về việc", "ko": " không", "đ/c": "địa chỉ"}
    for k, v in abbrs.items():
        text = text.replace(k, v)
    return text

def expand_standalone_letters(text):
    def _repl_letter(m):
        char = m.group(1).lower()
        if char in _letter_key_vi:
            return f" {_letter_key_vi[char]} "
        return m.group(0)
    
    return RE_STANDALONE_LETTER.sub(_repl_letter, text)

def normalize_urls(text):
    def _repl_url(m):
        url = m.group(0)
        res = []
        for char in url.lower():
            if char.isalnum():
                if char.isdigit():
                    res.append(n2w_single(char))
                else:
                    res.append(_vi_letter_names.get(char, char))
            elif char == '.': res.append('chấm')
            elif char == '/': res.append('xẹt')
            elif char == ':': res.append('hai chấm')
            elif char == '-': res.append('gạch ngang')
            elif char == '_': res.append('gạch dưới')
            elif char == '?': res.append('hỏi')
            elif char == '&': res.append('và')
            elif char == '=': res.append('bằng')
            else: res.append(char)
        return " ".join(res)

    return RE_URL.sub(_repl_url, text)

def normalize_slashes(text):
    def _repl(m):
        n1 = m.group(1)
        n2 = m.group(2)
        # If it's likely an address (first number is large)
        if len(n1) > 2 or int(n1) > 31:
            return f"{n2w(n1)} xẹt {n2w(n2)}"
        return f"{n2w(n1)} trên {n2w(n2)}"
    return RE_SLASH_NUMBER.sub(_repl, text)

def normalize_emails(text):
    def _repl_email(m):
        email = m.group(0)
        parts = email.split('@')
        if len(parts) != 2: return email

        user_part, domain_part = parts

        # User part: spell out
        user_norm = []
        for char in user_part.lower():
            if char.isalnum():
                if char.isdigit():
                    user_norm.append(n2w_single(char))
                else:
                    user_norm.append(_vi_letter_names.get(char, char))
            elif char == '.': user_norm.append('chấm')
            elif char == '_': user_norm.append('gạch dưới')
            elif char == '-': user_norm.append('gạch ngang')
            else: user_norm.append(char)

        # Domain part
        domain_part_lower = domain_part.lower()
        if domain_part_lower in _common_email_domains:
            domain_norm = _common_email_domains[domain_part_lower]
        else:
            domain_parts = domain_part.split('.')
            norm_domain_parts = []
            for dp in domain_parts:
                dp_norm = []
                for char in dp.lower():
                    if char.isalnum():
                        if char.isdigit():
                            dp_norm.append(n2w_single(char))
                        else:
                            dp_norm.append(_vi_letter_names.get(char, char))
                    else: dp_norm.append(char)
                norm_domain_parts.append(" ".join(dp_norm))
            domain_norm = " chấm ".join(norm_domain_parts)

        return " ".join(user_norm) + " a còng " + domain_norm

    return RE_EMAIL.sub(_repl_email, text)

WORD_LIKE_ACRONYMS = {"UNESCO", "NASA", "NATO", "ASEAN", "OPEC", "SARS", "FIFA", "UNIC", "RAM", "VRAM", "COVID", "IELTS", "STEM", "SWAT", "SEAL", "WASP", "COBOL", "BASIC", "OLED", "COVAX", "BRICS", "APEC", "VUCA", "PERMA", "DINK", "MENA", "EPIC", "OASIS", "BASE", "DART", "IDEA", "CHAOS", "SMART", "FANG"}

def normalize_acronyms(text):
    sentences = RE_SENTENCE_SPLIT.split(text)
    processed = []
    for i in range(0, len(sentences), 2):
        s = sentences[i]
        sep = sentences[i+1] if i+1 < len(sentences) else ""
        if not s:
            processed.append(sep)
            continue

        words = s.split()
        alpha_words = [w for w in words if any(c.isalpha() for c in w)]
        is_all_caps = len(alpha_words) > 0 and all(w.isupper() for w in alpha_words)

        if not is_all_caps:
            def _repl_acronym(m):
                word = m.group(0)
                if word.isdigit(): return word
                if word in WORD_LIKE_ACRONYMS:
                    return f"__START_EN__{word.lower()}__END_EN__"
                if any(c.isdigit() for c in word):
                    if word.upper() == "B2B":
                        return "__START_EN__b two b__END_EN__"
                    
                    res = []
                    for c in word.lower():
                        if c.isdigit():
                            res.append(n2w_single(c))
                        else:
                            res.append(_vi_letter_names.get(c, c))
                    return " ".join(res)

                spaced_word = " ".join(c.lower() for c in word if c.isalnum())
                if spaced_word:
                    return f"__START_EN__{spaced_word}__END_EN__"
                return word

            s = RE_ACRONYM.sub(_repl_acronym, s)

        processed.append(s + sep)
    return "".join(processed)

def expand_alphanumeric(text):
    def _repl(m):
        num = m.group(1)
        char = m.group(2).lower()
        if char in _letter_key_vi:
            pronunciation = _letter_key_vi[char]
            # Special case for roads (Quốc lộ 1D -> Quốc lộ 1 đê)
            if char == 'd' and ('quốc lộ' in text.lower() or 'ql' in text.lower()):
                pronunciation = 'đê'
            return f"{num} {pronunciation}"
        return m.group(0)
    return RE_ALPHANUMERIC.sub(_repl, text)

def expand_symbols(text):
    symbols_map = {
        '&': ' và ', '+': ' cộng ', '=': ' bằng ', '#': ' thăng ',
        '>': ' lớn hơn ', '<': ' nhỏ hơn ',
        '≥': ' lớn hơn hoặc bằng ', '≤': ' nhỏ hơn hoặc bằng ',
        '±': ' cộng trừ ', '≈': ' xấp xỉ '
    }
    for s, v in symbols_map.items():
        text = text.replace(s, v)
    return text

def expand_temperatures(text):
    text = RE_TEMP_C_NEG.sub(r'âm \1 độ xê', text)
    text = RE_TEMP_F_NEG.sub(r'âm \1 độ ép', text)
    text = RE_TEMP_C.sub(r'\1 độ xê', text)
    text = RE_TEMP_F.sub(r'\1 độ ép', text)
    text = RE_DEGREE.sub(' độ ', text)
    return text

def normalize_others(text):
    # 1. Expand acronym exceptions and basic patterns
    for pattern, v in _ACRONYMS_EXCEPTIONS_RE:
        text = pattern.sub(v, text)
    
    text = normalize_urls(text)
    text = normalize_emails(text)
    text = normalize_slashes(text)

    # 2. Expand Roman numerals and standalone letters
    text = RE_ROMAN_NUMBER.sub(expand_roman, text)
    text = RE_LETTER.sub(expand_letter, text)
    text = expand_alphanumeric(text)
    
    # 3. Clean quotes and symbols
    text = re.sub(r'["\'“”‘’]', '', text)
    text = expand_symbols(text)

    # 4. Expand units and measurements
    text = expand_compound_units(text)
    text = expand_measurement(text)
    text = expand_currency(text)
    
    # 5. Handle brackets and temperatures
    text = RE_BRACKETS.sub(r', \1, ', text)
    text = RE_STRIP_BRACKETS.sub(' ', text)
    text = expand_temperatures(text)

    # 6. Normalize acronyms
    text = normalize_acronyms(text)

    # 7. Expand version numbers (1.2.3 -> 1 chấm 2 chấm 3)
    text = RE_VERSION.sub(lambda m: ' chấm '.join(m.group(1).split('.')), text)

    # 8. Final cleanup
    text = RE_CLEAN_OTHERS.sub(' ', text)
    
    text = text.replace('__START_EN__', '<en>').replace('__END_EN__', '</en>')
    
    return text



import pytest
from vieneu_utils.cleaner.num2vi import n2w, n2w_single

def test_n2w_zero_padding():
    assert n2w("007") == "bảy"
    assert n2w("010") == "mười"
    assert n2w("000") == "không"

def test_n2w_large_scales():
    # 1,000,000,000,000 (1 trillion) - In VN usually "một nghìn tỷ"
    assert n2w("1000000000000") == "một nghìn tỷ"
    # 1,000,000,000,000,000 (1 quadrillion) - "một triệu tỷ"
    assert n2w("1000000000000000") == "một triệu tỷ"
    # 1,000,000,000,000,000,000 (1 quintillion) - "một tỷ tỷ"
    assert n2w("1000000000000000000") == "một tỷ tỷ"

def test_n2w_mot_vs_mot():
    assert n2w("21") == "hai mươi mốt"
    assert n2w("11") == "mười một"
    assert n2w("1") == "một"
    assert n2w("101") == "một trăm lẻ một"
    assert n2w("121") == "một trăm hai mươi mốt"

def test_n2w_lam_vs_nam():
    assert n2w("5") == "năm"
    assert n2w("15") == "mười lăm"
    assert n2w("25") == "hai mươi lăm"
    assert n2w("105") == "một trăm lẻ lăm"

def test_n2w_single_extended():
    assert n2w_single("0912") == "không chín một hai"
    assert n2w_single("+84912") == "không chín một hai"

def test_n2w_hundreds_internal():
    from vieneu_utils.cleaner.num2vi import n2w_hundreds
    assert n2w_hundreds("005") == "không trăm lẻ lăm"
    assert n2w_hundreds("015") == "không trăm mười lăm"
    assert n2w_hundreds("100") == "một trăm"



import re
import os
from typing import List, Optional
import numpy as np

# Pre-compile regex for splitting
RE_NEWLINE = re.compile(r"[\r\n]+")
RE_SENTENCE_END = re.compile(r"(?<=[\.\!\?\…])\s+")
RE_MINOR_PUNCT = re.compile(r"(?<=[\,\;\:\-\–\—])\s+")

def join_audio_chunks(chunks: List[np.ndarray], sr: int, silence_p: float = 0.0, crossfade_p: float = 0.0) -> np.ndarray:
    """
    Join audio chunks with optional silence padding and crossfading.

    Args:
        chunks: List of audio waveforms as numpy arrays.
        sr: Sample rate.
        silence_p: Duration of silence to insert between chunks (seconds).
        crossfade_p: Duration of crossfade between chunks (seconds).

    Returns:
        Joined audio waveform.
    """
    if not chunks:
        return np.array([], dtype=np.float32)
    if len(chunks) == 1:
        return chunks[0]
    
    silence_samples = int(sr * silence_p)
    crossfade_samples = int(sr * crossfade_p)
    
    final_wav = chunks[0]
    
    for i in range(1, len(chunks)):
        next_chunk = chunks[i]
        
        if silence_samples > 0:
            # 1. Add silence between chunks
            silence = np.zeros(silence_samples, dtype=np.float32)
            final_wav = np.concatenate([final_wav, silence, next_chunk])
        elif crossfade_samples > 0:
            # 2. Crossfade between chunks
            overlap = min(len(final_wav), len(next_chunk), crossfade_samples)
            if overlap > 0:
                fade_out = np.linspace(1.0, 0.0, overlap, dtype=np.float32)
                fade_in = np.linspace(0.0, 1.0, overlap, dtype=np.float32)
                
                blended = (final_wav[-overlap:] * fade_out + next_chunk[:overlap] * fade_in)
                final_wav = np.concatenate([
                    final_wav[:-overlap],
                    blended,
                    next_chunk[overlap:]
                ])
            else:
                final_wav = np.concatenate([final_wav, next_chunk])
        else:
            # 3. Simple concatenation
            final_wav = np.concatenate([final_wav, next_chunk])
            
    return final_wav

def split_text_into_chunks(text: str, max_chars: int = 256) -> List[str]:
    """
    Split raw text into chunks no longer than max_chars.

    Args:
        text: Input text to split.
        max_chars: Maximum number of characters per chunk.

    Returns:
        List of text chunks.
    """
    if not text:
        return []

    # 1. First split by newlines - each line/paragraph is handled independently
    paragraphs = RE_NEWLINE.split(text.strip())
    final_chunks: List[str] = []

    for para in paragraphs:
        para = para.strip()
        if not para:
            continue
            
        # 2. Split current paragraph into sentences
        sentences = RE_SENTENCE_END.split(para)
        
        buffer = ""
        for sentence in sentences:
            sentence = sentence.strip()
            if not sentence:
                continue
                
            # If sentence itself is longer than max_chars, we must split it by minor punctuation or words
            if len(sentence) > max_chars:
                # Flush buffer before handling a giant sentence
                if buffer:
                    final_chunks.append(buffer)
                    buffer = ""
                
                # Split giant sentence by minor punctuation (, ; : -)
                sub_parts = RE_MINOR_PUNCT.split(sentence)
                for part in sub_parts:
                    part = part.strip()
                    if not part: continue
                    
                    if len(buffer) + 1 + len(part) <= max_chars:
                        buffer = (buffer + " " + part) if buffer else part
                    else:
                        if buffer: final_chunks.append(buffer)
                        buffer = part
                        
                        # If even a sub-part is too long, split by spaces (words)
                        if len(buffer) > max_chars:
                            words = buffer.split()
                            current = ""
                            for word in words:
                                if current and len(current) + 1 + len(word) > max_chars:
                                    final_chunks.append(current)
                                    current = word
                                else:
                                    current = (current + " " + word) if current else word
                            buffer = current
            else:
                # Normal sentence: check if it fits in current buffer
                if buffer and len(buffer) + 1 + len(sentence) > max_chars:
                    final_chunks.append(buffer)
                    buffer = sentence
                else:
                    buffer = (buffer + " " + sentence) if buffer else sentence
        
        # End of paragraph: flush whatever is in buffer
        if buffer:
            final_chunks.append(buffer)
            buffer = ""

    return [c.strip() for c in final_chunks if c.strip()]

def env_bool(name: str, default: bool = False) -> bool:
    """Get boolean value from environment variable."""
    v = os.getenv(name)
    if v is None:
        return default
    return v.strip().lower() in ("1", "true", "yes", "y", "on")



import re
import unicodedata
from .cleaner import clean_vietnamese_text

class VietnameseTTSNormalizer:
    """
    A text normalizer for Vietnamese Text-to-Speech systems.
    Converts numbers, dates, units, and special characters into readable Vietnamese text.
    All core logic is implemented in the cleaner module.
    """
    
    def __init__(self) -> None:
        pass
    
    def normalize(self, text: str) -> str:
        """Main normalization pipeline with EN tag protection."""
        if not text:
            return ""

        # Pre-normalization: Ensure NFC format for Vietnamese characters
        text = unicodedata.normalize('NFC', text)

        # Step 1: Detect and protect EN tags
        en_contents = []
        placeholder_pattern = "___PROTECTED_EN_TAG_{}___"
        
        def extract_en(match):
            en_contents.append(match.group(0))
            return placeholder_pattern.format(len(en_contents) - 1)
        
        text = re.sub(r'<en>.*?</en>', extract_en, text, flags=re.IGNORECASE)
        
        # Step 2: Core Normalization
        text = clean_vietnamese_text(text)
        
        # Final cleanup - preserve newlines
        text = text.lower()
        text = re.sub(r'[ \t\xA0]+', ' ', text).strip()
        
        # Step 3: Restore EN tags
        for idx, en_content in enumerate(en_contents):
            placeholder = placeholder_pattern.format(idx).lower()
            text = text.replace(placeholder, en_content + ' ')
        
        # Final whitespace cleanup - preserve newlines
        text = re.sub(r'[ \t\xA0]+', ' ', text).strip()
        
        return text

if __name__ == "__main__":
    normalizer = VietnameseTTSNormalizer()
    
    test_texts = [
        "Lễ kỷ niệm 70 năm Chiến thắng Điện Biên Phủ (07/5/1954 - 07/5/2024).",
        "Phiên bản 1.0.4, tốc độ 60km/h, nhiệt độ 30°C.",
        "Số điện thoại: +84 912.345.678"
    ]
    
    for t in test_texts:
        print(f"In:  {t}")
        print(f"Out: {normalizer.normalize(t)}\n")


from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, Union, List, Dict, Any, Generator
import json
import torch
import numpy as np
import gc
import logging
from huggingface_hub import hf_hub_download
from vieneu_utils.normalize_text import VietnameseTTSNormalizer

# Configure logging
logger = logging.getLogger("Vieneu")

class BaseVieneuTTS(ABC):
    """
    Abstract base class for VieNeu-TTS implementations.
    Provides shared functionality for voice management and common operations.
    """

    def __init__(self):
        self.sample_rate = 24_000
        self.max_context = 2048
        self.hop_length = 480

        self.assets_dir = Path(__file__).parent / "assets"
        self._preset_voices: Dict[str, Any] = {}
        self._default_voice: Optional[str] = None
        self.normalizer = VietnameseTTSNormalizer()
        self._ref_phoneme_cache: Dict[str, str] = {}

        # Watermarker placeholder
        self.watermarker = None
        self._init_watermarker()

    def _init_watermarker(self) -> None:
        """Initialize optional audio watermarker."""
        try:
            import perth
            self.watermarker = perth.PerthImplicitWatermarker()
            logger.info("🔒 Audio watermarking initialized (Perth)")
        except (ImportError, AttributeError):
            self.watermarker = None

    def _load_voices(self, backbone_repo: Optional[str], hf_token: Optional[str] = None, clear_existing: bool = False) -> None:
        """Unified voice loading for Local and Remote paths."""
        if not backbone_repo:
            return

        path_obj = Path(backbone_repo)
        if path_obj.exists():
            # Local Path (Dir or File)
            if path_obj.is_dir():
                json_path = path_obj / "voices.json"
            else:
                json_path = path_obj.parent / "voices.json"

            if json_path.exists():
                self._load_voices_from_file(json_path, clear_existing=clear_existing)
            else:
                if clear_existing:
                     self._preset_voices.clear()
                logger.warning(f"Validation Warning: Local path '{backbone_repo}' missing 'voices.json'.")
                logger.warning(f"Falling back to Custom Voice Cloning mode.")
        else:
            # Remote Repo
            if clear_existing:
                self._preset_voices.clear()

            try:
                self._load_voices_from_repo(backbone_repo, hf_token)
            except Exception as e:
                logger.warning(f"Could not load voices from repo '{backbone_repo}': {e}")
                logger.warning(f"Falling back to Custom Voice Cloning mode.")

    def _load_voices_from_file(self, file_path: Path, clear_existing: bool = False) -> None:
        """Load voices from a local JSON file."""
        try:
            if not file_path.exists():
                logger.error(f"Voice file not found: {file_path}")
                return

            with open(file_path, 'r', encoding='utf-8') as f:
                try:
                    data = json.load(f)
                except json.JSONDecodeError as e:
                    logger.error(f"Invalid JSON in voice file {file_path}: {e}")
                    return

            if "presets" in data:
                if clear_existing:
                    self._preset_voices.clear()
                    logger.info("🧹 Cleared existing voices for replacement")

                # Merge into existing presets
                self._preset_voices.update(data["presets"])
                logger.info(f"📢 Loaded {len(data['presets'])} voices from {file_path.name}")

            # Update default voice if provided
            if "default_voice" in data and data["default_voice"]:
                self._default_voice = data["default_voice"]

        except Exception as e:
            logger.error(f"Failed to load voices from {file_path}: {e}")

    def _load_voices_from_repo(self, repo_id: str, hf_token: Optional[str] = None) -> None:
        """Download and load voices.json from a HuggingFace repo."""
        voices_file = None
        try:
            # 1. Try normal download (checks for updates from server)
            voices_file = hf_hub_download(
                repo_id=repo_id,
                filename="voices.json",
                token=hf_token,
                repo_type="model"
            )
        except Exception:
            # 2. Network error? Try to use cached version if available
            logger.warning(f"Network check failed for voices.json. Trying local cache...")
            try:
                voices_file = hf_hub_download(
                    repo_id=repo_id,
                    filename="voices.json",
                    token=hf_token,
                    repo_type="model",
                    local_files_only=True
                )
                logger.info(f"✅ Using cached voices.json")
            except Exception:
                # 3. No cache available either
                pass

        if voices_file:
            self._load_voices_from_file(Path(voices_file))
        else:
            logger.warning(f"Repository '{repo_id}' is missing 'voices.json'. Falling back to Custom Voice mode.")

    def list_preset_voices(self) -> List[tuple[str, str]]:
        """List available preset voices as (description, id)."""
        return [
            (v.get("description", k) if isinstance(v, dict) else str(v), k)
            for k, v in self._preset_voices.items()
        ]

    def get_preset_voice(self, voice_name: Optional[str] = None) -> Dict[str, Any]:
        """
        Get reference codes and text for a preset voice.

        Args:
            voice_name: Name of voice. If None, uses default_voice.

        Returns:
            dict: { 'codes': torch.Tensor, 'text': str }
        """
        if voice_name is None:
            voice_name = self._default_voice
            if voice_name is None:
                if self._preset_voices:
                    voice_name = next(iter(self._preset_voices))
                else:
                    raise ValueError("No voice specified and no preset voices available.")

        if voice_name not in self._preset_voices:
            raise ValueError(f"Voice '{voice_name}' not found. Available: {self.list_preset_voices()}")

        voice_data = self._preset_voices[voice_name]
        codes = voice_data["codes"]
        if isinstance(codes, list):
            codes = torch.tensor(codes, dtype=torch.long)

        return {"codes": codes, "text": voice_data["text"]}

    def get_ref_phonemes(self, ref_text: str) -> str:
        """
        Get phonemized version of reference text, using cache if available.
        """
        if ref_text not in self._ref_phoneme_cache:
            from vieneu_utils.phonemize_text import phonemize_with_dict
            self._ref_phoneme_cache[ref_text] = phonemize_with_dict(ref_text)
        return self._ref_phoneme_cache[ref_text]

    def save(self, audio: np.ndarray, output_path: Union[str, Path]) -> None:
        """Save audio waveform to a file."""
        import soundfile as sf
        sf.write(str(output_path), audio, self.sample_rate)

    def encode_reference(self, ref_audio_path: Union[str, Path]) -> torch.Tensor:
        """
        Encode reference audio to codes.

        Args:
            ref_audio_path: Path to the reference audio file.

        Returns:
            torch.Tensor: Encoded codes.
        """
        import librosa
        wav, _ = librosa.load(ref_audio_path, sr=16000, mono=True)
        wav_tensor = torch.from_numpy(wav).float().unsqueeze(0).unsqueeze(0)  # [1, 1, T]

        # Ensure device and dtype compatibility
        if hasattr(self.codec, "device"):
            wav_tensor = wav_tensor.to(self.codec.device)

        with torch.no_grad():
            ref_codes = self.codec.encode_code(audio_or_path=wav_tensor).squeeze(0).squeeze(0)
        return ref_codes

    def _decode(self, codes_str: str) -> np.ndarray:
        """
        Decode speech tokens to audio waveform.

        Args:
            codes_str: String containing speech tokens.

        Returns:
            np.ndarray: Decoded audio waveform.
        """
        from .utils import extract_speech_ids
        speech_ids = extract_speech_ids(codes_str)

        if len(speech_ids) == 0:
            raise ValueError("No valid speech tokens found in the output.")

        # Onnx decode
        if getattr(self, "_is_onnx_codec", False):
            codes = np.array(speech_ids, dtype=np.int32)[np.newaxis, np.newaxis, :]
            recon = self.codec.decode_code(codes)
        # Torch decode
        else:
            with torch.no_grad():
                codes = torch.tensor(speech_ids, dtype=torch.long)[None, None, :].to(
                    self.codec.device
                )
                recon = self.codec.decode_code(codes).cpu().numpy()

        return recon[0, 0, :]

    def _resolve_ref_voice(
        self,
        voice: Optional[Dict[str, Any]] = None,
        ref_audio: Optional[Union[str, Path]] = None,
        ref_codes: Optional[Union[np.ndarray, torch.Tensor]] = None,
        ref_text: Optional[str] = None
    ) -> tuple[Union[np.ndarray, torch.Tensor], str]:
        """Resolve reference voice codes and text."""
        if voice is not None:
            ref_codes = voice.get('codes', ref_codes)
            ref_text = voice.get('text', ref_text)

        if ref_audio is not None and ref_codes is None:
            ref_codes = self.encode_reference(ref_audio)
        elif self._default_voice and (ref_codes is None or ref_text is None):
            try:
                voice_data = self.get_preset_voice(None)
                ref_codes = voice_data['codes']
                ref_text = voice_data['text']
            except Exception:
                pass

        if ref_codes is None or ref_text is None:
            raise ValueError("Must provide either 'voice' dict or both 'ref_codes' and 'ref_text'.")

        return ref_codes, ref_text

    def _apply_watermark(self, wav: np.ndarray) -> np.ndarray:
        """Apply watermark to audio if enabled."""
        if self.watermarker:
            return self.watermarker.apply_watermark(wav, sample_rate=self.sample_rate)
        return wav

    @abstractmethod
    def infer(self, text: str, **kwargs: Any) -> np.ndarray:
        """Main inference method for single text."""
        pass

    @abstractmethod
    def infer_batch(self, texts: List[str], apply_watermark: bool = True, **kwargs: Any) -> List[np.ndarray]:
        """Main inference method for batch processing."""
        pass

    def close(self) -> None:
        """Release resources."""
        pass

    def __enter__(self) -> 'BaseVieneuTTS':
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.close()

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass



import pytest
from vieneu_utils.normalize_text import VietnameseTTSNormalizer

@pytest.fixture
def normalizer():
    return VietnameseTTSNormalizer()

# Combined test cases from multiple categories
TEST_CASES = [
    # ─── 1. SỐ THÔNG THƯỜNG ────────────────────────────────────────────────────
    ("0",        "không"),
    ("1",        "một"),
    ("10",       "mười"),
    ("11",       "mười một"),
    ("21",       "hai mươi mốt"),
    ("100",      "một trăm"),
    ("1000",     "một nghìn"),
    ("1001",     "một nghìn không trăm lẻ một"),
    ("1000000",  "một triệu"),
    ("1234567",  "một triệu hai trăm ba mươi bốn nghìn năm trăm sáu mươi bảy"),

    # ─── 2. SỐ THẬP PHÂN / SỐ CÓ DẤU PHÂN CÁCH ──────────────────────────────
    ("1.000",    "một nghìn"),
    ("1.000.000","một triệu"),
    ("3,14",     "ba phẩy mười bốn"),
    ("1.3",      "một chấm ba"),

    # ─── 3. SỐ ĐIỆN THOẠI ────────────────────────────────────────────────────
    ("0912345678",    "không chín một hai ba bốn năm sáu bảy tám"),
    ("+84912345678",  "cộng tám bốn chín một hai ba bốn năm sáu bảy tám"),

    # ─── 4. SỐ THỨ TỰ ─────────────────────────────────────────────────────────
    ("thứ 1",  "thứ nhất"),
    ("thứ 4",  "thứ tư"),
    ("thứ 5",  "thứ năm"),
    ("hạng 1", "hạng nhất"),

    # ─── 5. PHÉP NHÂN ─────────────────────────────────────────────────────────
    ("3 x 4",  "ba nhân bốn"),
    ("10 x 20","mười nhân hai mươi"),

    # ─── 6. NGÀY THÁNG ─────────────────────────────────────────────────────────
    ("21/02/2025", "ngày hai mươi mốt tháng hai năm hai nghìn không trăm hai mươi lăm"),
    ("01-01-2024", "ngày một tháng một năm hai nghìn không trăm hai mươi bốn"),
    ("31.12.2023", "ngày ba mươi mốt tháng mười hai năm hai nghìn không trăm hai mươi ba"),
    ("31.12.1997", "ngày ba mươi mốt tháng mười hai năm một nghìn chín trăm chín mươi bảy"),
    ("21/02", "ngày hai mươi mốt tháng hai"),
    ("01/12", "ngày một tháng mười hai"),
    ("02/2025", "tháng hai năm hai nghìn không trăm hai mươi lăm"),
    ("12/2024", "tháng mười hai năm hai nghìn không trăm hai mươi bốn"),
    ("32/01", "ba mươi hai xẹt không một"),
    ("01/13", "không một trên mười ba"),
    ("tháng 3/2026", "tháng ba năm hai nghìn không trăm hai mươi sáu"),

    # ─── 7. THỜI GIAN ─────────────────────────────────────────────────────────
    ("14h30",   "mười bốn giờ ba mươi phút"),
    ("8h05",    "tám giờ không năm phút"),
    ("0h00",    "không giờ không phút"),
    ("23:59",   "hai mươi ba giờ năm mươi chín phút"),
    ("12:00:00","mười hai giờ không phút không giây"),
    ("10:20 phút", "mười giờ hai mươi phút"),
    ("12:00:00 giây", "mười hai giờ không phút không giây"),

    # ─── 8. TIỀN TỆ ──────────────────────────────────────────────────────────
    ("100$",   "một trăm đô la Mỹ"),
    ("$50",    "năm mươi đô la Mỹ"),
    ("200 USD","hai trăm đô la Mỹ"),
    ("500 VND","năm trăm đồng"),
    ("50 euro","năm mươi ơ rô"),
    ("1000đ",  "một nghìn đồng"),
    ("75%",    "bảy mươi lăm phần trăm"),
    ("15,4% xuống còn 8,3%", "mười lăm phẩy bốn phần trăm xuống còn tám phẩy ba phần trăm"),
    ("370 tỷ USD", "ba trăm bảy mươi tỷ đô la Mỹ"),
    ("5 triệu VND", "năm triệu đồng"),
    ("10 nghìn USD", "mười nghìn đô la Mỹ"),
    ("8,92 tỷ USD", "tám phẩy chín mươi hai tỷ đô la Mỹ"),

    # ─── 9. ĐƠN VỊ ĐO LƯỜNG ─────────────────────────────────────────────────
    ("50km",  "năm mươi ki lô mét"),
    ("100m",  "một trăm mét"),
    ("30cm",  "ba mươi xen ti mét"),
    ("5mm",   "năm mi li mét"),
    ("75kg",  "bảy mươi lăm ki lô gam"),
    ("500g",  "năm trăm gam"),
    ("250ml", "hai trăm năm mươi mi li lít"),
    ("2l",    "hai lít"),
    ("10ha",  "mười héc ta"),
    ("50m2",  "năm mươi mét vuông"),
    ("20m3",  "hai mươi mét khối"),
    ("300.000km", "ba trăm nghìn ki lô mét"),
    ("5 triệu km", "năm triệu ki lô mét"),
    ("1,5 ha", "một phẩy năm héc ta"),
    ("1.5 ha", "một chấm năm héc ta"),
    ("AN/ASQ", "an trên asq"),

    # ─── 10. KHOẢNG / DÃY SỐ ──────────────────────────────────────────────────
    ("700-900", "bảy trăm đến chín trăm"),
    ("0,5-0,9", "không phẩy năm đến không phẩy chín"),

    # ─── 11. SỐ LA MÃ ────────────────────────────────────────────────────────
    ("Thế kỷ XXI",  "thế kỷ hai mươi mốt"),
    ("Chương IV",   "chương bốn"),
    ("Hồi IX",      "hồi chín"),
    ("Phần III",    "phần ba"),
    ("Thế kỷ XX",   "thế kỷ hai mươi"),

    # ─── 12. CHỮ CÁI ─────────────────────────────────────────────────────────
    ("ký tự A",     "ký tự a"),
    ("chữ B",       "chữ bê"),
    ("ký tự 'C'",   "ký tự xê"),
    ("chữ cái Z",   "chữ cái dét"),
    ("kí tự w",     "kí tự vê kép"),

    # ─── 13. TỪ VIẾT TẮT TIẾNG VIỆT ─────────────────────────────────────────
    ("UBND",  "uỷ ban nhân dân"),
    ("TP.HCM","thành phố hồ chí minh"),
    ("TPHCM", "thành phố hồ chí minh"),
    ("CSGT",  "cảnh sát giao thông"),
    ("LHQ",   "liên hợp quốc"),
    ("CLB",   "câu lạc bộ"),
    ("HLV",   "huấn luyện viên"),
    ("TS",    "tiến sĩ"),
    ("GS",    "giáo sư"),
    ("THPT",  "trung học phổ thông"),
    ("THCS",  "trung học cơ sở"),

    # ─── 14. THẺ TIẾNG ANH (EN TAG) ─────────────────────────────────────────
    ("<en>Hello</en>",                              "<en>Hello</en>"),
    ("<en>Hello 123</en>",                          "<en>Hello 123</en>"),
    ("Xin chào <en>Good morning</en>",              "xin chào <en>Good morning</en>"),
    ("Ngày 21/02 <en>February 21</en>",             "ngày hai mươi mốt tháng hai <en>February 21</en>"),
    ("<en>AI</en> là trí tuệ nhân tạo",             "<en>AI</en> là trí tuệ nhân tạo"),

    # ─── 15. DẤU CÂU ─────────────────────────────────────────────────────────
    ("A & B",              "a và bê"),
    ("A + B",              "a cộng bê"),
    ("A = B",              "a bằng bê"),
    ("#1",                 "thăng một"),
    ("(text in brackets)", "text in brackets"),
    ("[text in brackets]", "text in brackets"),
    ("(giờ Mỹ)", "giờ mỹ"),
    ("hiệu lực từ 0h01 (giờ Mỹ), trong vòng", "hiệu lực từ không giờ không một phút, giờ mỹ, trong vòng"),
    ("hiệu lực từ 0h01 (giờ Mỹ) trong vòng", "hiệu lực từ không giờ không một phút, giờ mỹ, trong vòng"),
    ("kết thúc (0h01).", "kết thúc, không giờ không một phút."),
    ("chỉ số là 7,05 - đường huyết là 1.8", "chỉ số là bảy phẩy không năm, đường huyết là một chấm tám"),

    # ─── 16. CẤU TRÚC VĂN BẢN ──────────────────────────────────────────────
    ("Đoạn 1.\nĐoạn 2.", "đoạn một.\nđoạn hai."),
    ("\n12 tiêm kích", "\nmười hai tiêm kích"),

    # ─── 17. VIẾT TẮT ĐƠN GIẢN ──────────────────────────────────────────────
    ("v.v",  "vân vân"),
    ("v/v",  "về việc"),
    ("ko",   "không"),
    ("đ/c",  "địa chỉ"),

    # ─── 18. TRƯỜNG HỢP HỖN HỢP ──────────────────────────────────────────────
    ("Ngày 21/02/2025 lúc 14h30, giá vàng đạt 100$ tại TPHCM",
     "ngày hai mươi mốt tháng hai năm hai nghìn không trăm hai mươi lăm lúc mười bốn giờ ba mươi phút, giá vàng đạt một trăm đô la Mỹ tại thành phố hồ chí minh"),
    ("Thế kỷ XXI chứng kiến sự phát triển của <en>AI</en> và vũ trụ học",
     "thế kỷ hai mươi mốt chứng kiến sự phát triển của <en>AI</en> và vũ trụ học"),
    ("Đề án 06 và Chỉ thị 04", "đề án không sáu và chỉ thị không bốn"),

    # ─── 19. AN TOÀN (KHÔNG NHẦM LẪN) ──────────────────────────────────────────
    ("Anh M. đi bộ", "anh mờ đi bộ"),
    ("Vitamin G", "vitamin gờ"),
    ("L. là tên riêng", "lờ là tên riêng"),
    ("5m chiều dài", "năm mét chiều dài"),
    ("Đơn vị km", "đơn vị ki lô mét"),

    # ─── 20. EMAIL ───────────────────────────────────────────────────────────
    ("Liên hệ qua email pnnbao@gmail.com nhé.", "liên hệ qua email phê nờ nờ bê a ô a còng gờ meo chấm com nhé."),
    ("Email: contact@example.com", "email: xê ô nờ tê a xê tê a còng e ích a mờ phê lờ e chấm xê ô mờ"),

    # ─── 21. VIẾT TẮT ALPHANUMERIC (ENGLISH STYLE) ──────────────────────────
    ("Mô hình B2B rất phổ biến.", "mô hình <en>b two b</en> rất phổ biến."),
    ("Tôi dùng camera K3.", "tôi dùng camera ca ba."),
    ("Mã số A1B.", "mã số a một bê."),
    ("Tôi đang học về AI.", "tôi đang học về <en>a i</en>."),
    ("Dự án VYE.", "dự án <en>v y e</en>."),

    # ─── 22. PHÂN BIỆT HOA THƯỜNG TRONG CÂU ─────────────────────────────────
    ("TÔI ĐI HỌC", "tôi đi học"),
    ("Chào mừng bạn đến với CTY.", "chào mừng bạn đến với <en>c t y</en>."),

    # ─── 23. TOÀN DIỆN (CẢI TIẾN MỚI) ──────────────────────────────────────────
    # URLs
    ("Truy cập https://vieneu.io để biết thêm chi tiết.", "truy cập hát tê tê phê ét hai chấm xẹt xẹt vờ i e nờ e u chấm i ô để biết thêm chi tiết."),
    ("Website www.google.com rất hữu ích.", "website vê kép vê kép vê kép chấm gờ ô ô gờ lờ e chấm xê ô mờ rất hữu ích."),

    # Slashes / Địa chỉ
    ("Địa chỉ nhà tôi là 123/4 đường Nguyễn Trãi.", "địa chỉ nhà tôi là một trăm hai mươi ba xẹt bốn đường nguyễn trãi."),
    ("Tỷ lệ là 100/2.", "tỷ lệ là một trăm xẹt hai."),

    # Ký hiệu toán học
    ("Nếu x > 5 và y ≤ 10 thì xấp xỉ ≈ 0.", "nếu ích lớn hơn năm và y nhỏ hơn hoặc bằng mười thì xấp xỉ xấp xỉ không."),
    ("Nhiệt độ là 30°C ± 2°C.", "nhiệt độ là ba mươi độ xê cộng trừ hai độ xê."),
    ("Biểu thức ≥ 10.", "biểu thức lớn hơn hoặc bằng mười."),

    # Đơn vị đo lường mở rộng
    ("Dung lượng 16GB.", "dung lượng mười sáu gi ga bai."),
    ("File nặng 50MB.", "file nặng năm mươi mê ga bai."),
    ("Ổ cứng 1TB.", "ổ cứng một tê ra bai."),
    ("Căn hộ 75sqm.", "căn hộ bảy mươi lăm mét vuông."),
    ("Bể bơi 100cum.", "bể bơi một trăm mét khối."),
    ("Âm thanh 80db.", "âm thanh tám mươi đê xi ben."),
    ("Trọng lượng 10lb.", "trọng lượng mười pao."),
    ("Màn hình 24in.", "màn hình hai mươi bốn ins."),
    ("Độ phân giải 300dpi.", "độ phân giải ba trăm đê phê i"),

    # Emails mở rộng
    ("Email công việc: admin@fpt.vn", "email công việc: a dê mờ i nờ a còng ép phê tê chấm vê nờ"),
    ("Liên hệ hotmail: test@hotmail.com", "liên hệ hotmail: tê e ét tê a còng hót meo chấm com"),

    # Redundant expansion (symbol + unit)
    ("#1kg", "thăng một ki lô gam"),

    # ─── 24. CÂU TEST THỰC TẾ ──────────────────────────────────────────────────
    ("Ông Lưu Trung Thái, Chủ tịch HĐQT MB cho biết, vốn hóa của ngân hàng đã tăng gần 10 lần kể từ năm 2017, đạt khoảng 8,5 tỷ USD, tạo nền tảng cho mục tiêu 10 tỷ USD vào năm 2027.",
     "ông lưu trung thái, chủ tịch hđqt <en>m b</en> cho biết, vốn hóa của ngân hàng đã tăng gần mười lần kể từ năm hai nghìn không trăm mười bảy, đạt khoảng tám phẩy năm tỷ đô la mỹ, tạo nền tảng cho mục tiêu mười tỷ đô la mỹ vào năm hai nghìn không trăm hai mươi bảy."),

    # ─── 25. THÊM MỚI (CẢI THIỆN) ──────────────────────────────────────────
    ("Số tiền là 1.000.000.000.000 đồng.", "số tiền là một nghìn tỷ đồng."),
    ("“Lời chào cao hơn mâm cỗ”", "lời chào cao hơn mâm cỗ"),
    ("‘Trân trọng’", "trân trọng"),
    ("Chào <en>world</en> xinh đẹp", "chào <en>world</en> xinh đẹp"),
]

@pytest.mark.parametrize("input_text, expected", TEST_CASES)
def test_normalize(normalizer, input_text, expected):
    actual = normalizer.normalize(input_text)
    # Clean up whitespace for comparison
    actual_clean = " ".join(actual.split()).lower()
    expected_clean = " ".join(expected.split()).lower()
    
    # Special handling for brackets/punctuation to match expectation
    # Note: brackets are sometimes kept or removed depending on normalizer state
    # We follow the user provided expected strings
    assert actual_clean == expected_clean



import re
from .num2vi import n2w, n2w_single
from .symbols import vietnamese_set

_normal_number_re        = r"[\d]+"
_float_number_re         = r"[\d]+[,]{1}[\d]+"
_number_with_one_dot     = r"[\d]+[.]{1}[\d]{3}"
_number_with_two_dot     = r"[\d]+[.]{1}[\d]{3}[.]{1}[\d]{3}"
_number_with_three_dot   = r"[\d]+[.]{1}[\d]{3}[.]{1}[\d]{3}[.]{1}[\d]{3}"
_number_with_one_space   = r"[\d]+[\s]{1}[\d]{3}"
_number_with_two_space   = r"[\d]+[\s]{1}[\d]{3}[\s]{1}[\d]{3}"
_number_with_three_space = r"[\d]+[\s]{1}[\d]{3}[\s]{1}[\d]{3}[\s]{1}[\d]{3}"

_number_combined = (
    r"("
    + _float_number_re + "|"
    + _number_with_three_dot + "|"
    + _number_with_two_dot + "|"
    + _number_with_one_dot + "|"
    + _number_with_three_space + "|"
    + _number_with_two_space + "|"
    + _number_with_one_space + "|"
    + _normal_number_re
    + r")"
)

# Compiled Regular Expressions
RE_NUMBER = re.compile(r"(\D)(-{1})?" + _number_combined)
RE_NUMBER_START = re.compile(r"^(-{1})?" + _number_combined, re.MULTILINE)
RE_MULTIPLY = re.compile(r"(" + _normal_number_re + r")(x|\sx\s)(" + _normal_number_re + r")")
RE_ORDINAL = re.compile(r"(thứ|hạng)(\s)(1|4)")
RE_PHONE = re.compile(r"((\+84|84|0|0084)(3|5|7|8|9)[0-9]{8})")
RE_DOT_SEP = re.compile(r"\d+(\.\d{3})+")

def _normalize_dot_sep(number: str) -> str:
    if RE_DOT_SEP.fullmatch(number):
        return number.replace(".", "")
    return number

def _num_to_words(number: str, negative: bool = False) -> str:
    number = _normalize_dot_sep(number).replace(" ", "")
    if "," in number:
        parts = number.split(",")
        return n2w(parts[0]) + " phẩy " + n2w(parts[1])
    elif negative:
        return "âm " + n2w(number)
    return n2w(number)

def _expand_number(match):
    prefix, negative_symbol, number = match.groups(0)
    negative = (negative_symbol == "-")
    word = _num_to_words(number, negative)
    prefix_str = "" if prefix in (0, None) else prefix
    return prefix_str + " " + word + " "

def _expand_number_start(match):
    negative_symbol, number = match.groups()
    negative = (negative_symbol == "-")
    return _num_to_words(number, negative) + " "

def _expand_phone(match):
    return n2w_single(match.group(0).strip())

def _expand_ordinal(match):
    prefix, space, number = match.groups(0)
    if number == "1": return prefix + space + "nhất"
    if number == "4": return prefix + space + "tư"
    return prefix + space + n2w(number)

def _expand_multiply_number(match):
    n1, _, n2 = match.groups(0)
    return n2w(n1) + " nhân " + n2w(n2)

def normalize_number_vi(text):
    text = RE_ORDINAL.sub(_expand_ordinal, text)
    text = RE_MULTIPLY.sub(_expand_multiply_number, text)
    text = RE_PHONE.sub(_expand_phone, text)
    # 1. Start of string OR start of line (handling newlines preserved by normalization)
    text = RE_NUMBER_START.sub(_expand_number_start, text)
    # 2. Anywhere else (preceded by a non-digit)
    text = RE_NUMBER.sub(_expand_number, text)
    return text
